import logging

from django.db.models.signals import post_save
from django.dispatch import receiver

from hub20.apps.blockchain.models import Chain

from .models import EthereumToken

logger = logging.getLogger(__name__)


@receiver(post_save, sender=Chain)
def on_chain_created_create_native_token(sender, **kw):
    if kw["created"]:
        chain = kw["instance"]
        EthereumToken.make_native(chain)



from decimal import Decimal

from django.test import TestCase
from django.urls import reverse
from rest_framework.test import APIClient

from hub20.apps.blockchain.factories import FAKER
from hub20.apps.core import factories
from hub20.apps.core.factories import (
    Erc20TokenBlockchainPaymentFactory,
    Erc20TokenPaymentConfirmationFactory,
)
from hub20.apps.ethereum_money.factories import Erc20TokenAmountFactory, Erc20TokenFactory


class StoreViewTestCase(TestCase):
    def setUp(self):
        self.client = APIClient()
        self.store = factories.StoreFactory()

    def test_anonymous_user_can_see_store(self):
        url = reverse("store-detail", kwargs={"pk": self.store.pk})
        response = self.client.get(url)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.data["id"], str(self.store.id))


class UserStoreViewTestCase(TestCase):
    def setUp(self):
        self.client = APIClient()
        self.store = factories.StoreFactory()

    def test_anonymous_user_can_not_list_stores(self):
        url = reverse("user-store-list")
        response = self.client.get(url)
        self.assertEqual(response.status_code, 401)

    def test_store_owner_can_see_store(self):
        url = reverse("store-detail", kwargs={"pk": self.store.pk})
        self.client.force_authenticate(user=self.store.owner)
        self.client.get(url)
        response = self.client.get(url)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.data["id"], str(self.store.id))

    def test_non_owner_can_not_see_store(self):
        url = reverse("user-store-detail", kwargs={"pk": self.store.pk})
        another_user = factories.UserFactory()
        self.client.force_authenticate(user=another_user)

        response = self.client.get(url)
        self.assertEqual(response.status_code, 403)


class UserViewTestCase(TestCase):
    def setUp(self):
        self.superuser = factories.UserFactory(is_superuser=True, is_staff=True)
        self.staff_user = factories.UserFactory(is_staff=True)
        self.inactive_user = factories.UserFactory(is_active=False)
        self.client = APIClient()
        self.url = reverse("users-list")

    def test_search_shows_only_active_users(self):
        regular_username = "one_regular_user"
        active_user = factories.UserFactory(username=regular_username)
        self.client.force_authenticate(user=active_user)

        response = self.client.get(self.url)
        self.assertEqual(response.status_code, 200)

        self.assertEqual(len(response.data), 1)
        self.assertEqual(response.data[0]["username"], regular_username)

    def test_filter_query(self):
        one_user = factories.UserFactory(username="one_user", email="hub20@one.example.com")
        factories.UserFactory(username="another_user", email="hub20@another.example.com")

        self.client.force_authenticate(user=one_user)

        one_query_response = self.client.get(self.url, {"search": "one"})
        self.assertEqual(len(one_query_response.data), 1)

        another_query_response = self.client.get(self.url, {"search": "another"})
        self.assertEqual(len(another_query_response.data), 1)

        email_query_response = self.client.get(self.url, {"search": "hub20"})
        self.assertEqual(len(email_query_response.data), 2)


class TokenBalanceViewTestCase(TestCase):
    def setUp(self):
        self.token = Erc20TokenFactory()
        self.user = factories.UserFactory()
        self.client = APIClient()
        self.client.force_authenticate(user=self.user)

        Erc20TokenPaymentConfirmationFactory(
            payment__route__deposit__user=self.user,
            payment__currency=self.token,
            payment__amount=10,
        )

    def test_balance_list_includes_token(self):
        response = self.client.get(reverse("balance-list"))
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(response.data), 1)

    def test_balance_view(self):
        response = self.client.get(
            reverse(
                "balance-detail",
                kwargs={"chain_id": self.token.chain_id, "address": self.token.address},
            )
        )
        self.assertEqual(response.status_code, 200)
        self.assertEqual(Decimal(response.data["amount"]), 10)


class TransferTestCase(TestCase):
    def setUp(self):
        self.user = factories.UserFactory()
        self.client = APIClient()
        self.client.force_authenticate(user=self.user)
        self.token = Erc20TokenFactory()
        self.target_address = FAKER.ethereum_address()

    def test_no_balance_returns_error(self):
        response = self.client.post(
            reverse("transfer-list"),
            {
                "address": self.target_address,
                "amount": 10,
                "token": reverse(
                    "token-detail",
                    kwargs=dict(address=self.token.address, chain_id=self.token.chain_id),
                ),
            },
        )
        self.assertEqual(response.status_code, 400)

        self.assertTrue("non_field_errors" in response.data.keys())
        self.assertEqual(len(response.data["non_field_errors"]), 1)

        error_details = response.data["non_field_errors"][0]
        self.assertEqual(error_details.code, "invalid")

    def test_insufficient_balance_returns_error(self):
        TRANSFER_AMOUNT = 10

        Erc20TokenPaymentConfirmationFactory(
            payment__route__deposit__user=self.user,
            payment__currency=self.token,
            payment__amount=TRANSFER_AMOUNT / 2,
        )

        response = self.client.post(
            reverse("transfer-list"),
            {
                "address": self.target_address,
                "amount": TRANSFER_AMOUNT,
                "token": reverse(
                    "token-detail",
                    kwargs=dict(address=self.token.address, chain_id=self.token.chain_id),
                ),
            },
        )
        self.assertEqual(response.status_code, 400)

        self.assertTrue("non_field_errors" in response.data.keys())
        self.assertEqual(len(response.data["non_field_errors"]), 1)

        error_details = response.data["non_field_errors"][0]
        self.assertEqual(error_details.code, "insufficient")


class CheckoutViewTestCase(TestCase):
    def setUp(self):
        self.token = Erc20TokenFactory()
        self.store = factories.StoreFactory(accepted_token_list__tokens=[self.token])

    def test_can_create_checkout_via_api(self):
        amount = Erc20TokenAmountFactory(currency=self.token)

        url = reverse("checkout-list")
        post_data = {
            "amount": amount.amount,
            "token": reverse(
                "token-detail",
                kwargs=dict(address=amount.currency.address, chain_id=amount.currency.chain_id),
            ),
            "store": self.store.id,
            "external_identifier": "API Test",
        }

        response = self.client.post(url, post_data)
        self.assertEqual(response.status_code, 201, response.data)

    def test_can_not_delete_checkout(self):
        checkout = factories.CheckoutFactory(store=self.store)
        url = reverse("checkout-detail", kwargs={"pk": checkout.pk})
        response = self.client.delete(url)
        self.assertEqual(response.status_code, 405)

    def test_payment_serializer(self):
        checkout = factories.CheckoutFactory(store=self.store)
        route = checkout.routes.select_subclasses().first()

        Erc20TokenBlockchainPaymentFactory(route=route)

        url = reverse("checkout-detail", kwargs={"pk": checkout.pk})
        response = self.client.get(url)

        self.assertEqual(len(response.data["payments"]), 1)

        payment = response.data["payments"][0]

        self.assertTrue("transaction" in payment)
        self.assertTrue("block" in payment)


__all__ = ["StoreViewTestCase", "CheckoutViewTestCase"]



import logging

from hub20.apps.blockchain.factories import TransactionFactory
from hub20.apps.blockchain.models import Transaction
from hub20.apps.blockchain.typing import EthereumAccount_T
from hub20.apps.ethereum_money.factories import Erc20TransferFactory
from hub20.apps.ethereum_money.models import EthereumTokenAmount
from hub20.apps.ethereum_money.signals import incoming_transfer_mined

logger = logging.getLogger(__name__)


def add_eth_to_account(account: EthereumAccount_T, amount: EthereumTokenAmount):
    tx = TransactionFactory(to_address=account.address)
    account.transactions.add(tx)
    incoming_transfer_mined.send(
        sender=Transaction,
        transaction=tx,
        amount=amount,
        account=account,
        address=tx.from_address,
    )
    return tx


def add_token_to_account(account: EthereumAccount_T, amount: EthereumTokenAmount):
    logging.debug(f"Creating Transfer transaction of {amount} to {account.address}")

    tx = Erc20TransferFactory(recipient=account.address, amount=amount)
    account.transactions.add(tx)
    incoming_transfer_mined.send(
        sender=Transaction,
        transaction=tx,
        amount=amount,
        account=account,
        address=tx.from_address,
    )
    return tx



# Generated by Django 4.0 on 2022-01-29 00:43

import django.contrib.postgres.fields.ranges
from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone
import hub20.apps.blockchain.fields
import hub20.apps.core.models.payments
import hub20.apps.ethereum_money.fields
import model_utils.fields
import uuid


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('ethereum_money', '0001_initial'),
        ('contenttypes', '0002_remove_content_type_name'),
        ('blockchain', '0001_initial'),
        ('auth', '0012_alter_user_first_name_max_length'),
        ('raiden', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='Book',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('owner_id', models.PositiveIntegerField()),
                ('owner_type', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='contenttypes.contenttype')),
                ('token', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, related_name='books', to='ethereum_money.ethereumtoken')),
            ],
            options={
                'unique_together': {('token', 'owner_type', 'owner_id')},
            },
        ),
        migrations.CreateModel(
            name='Deposit',
            fields=[
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified')),
                ('id', models.UUIDField(default=uuid.uuid4, primary_key=True, serialize=False)),
                ('session_key', models.SlugField(null=True)),
                ('currency', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, to='ethereum_money.ethereumtoken')),
                ('user', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='auth.user')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='ExternalAddressAccount',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('address', hub20.apps.blockchain.fields.EthereumAddressField(unique=True)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='Payment',
            fields=[
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified')),
                ('amount', hub20.apps.ethereum_money.fields.EthereumTokenAmountField(decimal_places=18, max_digits=32)),
                ('id', models.UUIDField(default=uuid.uuid4, primary_key=True, serialize=False)),
                ('currency', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, to='ethereum_money.ethereumtoken')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='PaymentRoute',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified')),
                ('deposit', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='routes', to='core.deposit')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='Store',
            fields=[
                ('id', models.UUIDField(default=uuid.uuid4, primary_key=True, serialize=False)),
                ('name', models.CharField(max_length=300)),
                ('url', models.URLField(help_text='URL for your store public site or information page')),
                ('checkout_webhook_url', models.URLField(help_text='URL to receive checkout updates', null=True)),
                ('accepted_token_list', models.ForeignKey(help_text='The list of tokens that will be accepted for payment', null=True, on_delete=django.db.models.deletion.SET_NULL, to='ethereum_money.usertokenlist')),
                ('owner', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='auth.user')),
            ],
        ),
        migrations.CreateModel(
            name='Transfer',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified')),
                ('amount', hub20.apps.ethereum_money.fields.EthereumTokenAmountField(decimal_places=18, max_digits=32)),
                ('address', hub20.apps.blockchain.fields.EthereumAddressField(db_index=True, null=True)),
                ('memo', models.TextField(blank=True, null=True)),
                ('identifier', models.CharField(blank=True, max_length=300, null=True)),
                ('execute_on', models.DateTimeField(auto_now_add=True)),
                ('currency', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, to='ethereum_money.ethereumtoken')),
                ('receiver', models.ForeignKey(null=True, on_delete=django.db.models.deletion.PROTECT, related_name='transfers_received', to='auth.user')),
                ('sender', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, related_name='transfers_sent', to='auth.user')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='TransferExecution',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified')),
                ('transfer', models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, related_name='execution', to='core.transfer')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='InternalPayment',
            fields=[
                ('payment_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='core.payment')),
                ('memo', models.TextField(blank=True, null=True)),
            ],
            options={
                'abstract': False,
            },
            bases=('core.payment',),
        ),
        migrations.CreateModel(
            name='InternalPaymentRoute',
            fields=[
                ('paymentroute_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='core.paymentroute')),
            ],
            options={
                'abstract': False,
            },
            bases=('core.paymentroute',),
        ),
        migrations.CreateModel(
            name='PaymentOrder',
            fields=[
                ('deposit_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='core.deposit')),
                ('amount', hub20.apps.ethereum_money.fields.EthereumTokenAmountField(decimal_places=18, max_digits=32)),
            ],
            options={
                'abstract': False,
            },
            bases=('core.deposit', models.Model),
        ),
        migrations.CreateModel(
            name='WalletAccount',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('account', models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, related_name='onchain_account', to='blockchain.baseethereumaccount')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='UserPreferences',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('tokens', models.ManyToManyField(to='ethereum_money.EthereumToken')),
                ('user', models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, related_name='preferences', to='auth.user')),
            ],
        ),
        migrations.CreateModel(
            name='UserAccount',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('user', models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, related_name='account', to='auth.user')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='Treasury',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('chain', models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, to='blockchain.chain')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='TransferReceipt',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified')),
                ('network', models.CharField(choices=[('blockchain', 'blockchain'), ('raiden', 'raiden'), ('internal', 'internal')], max_length=64)),
                ('identifier', models.CharField(max_length=300, null=True)),
                ('transfer', models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, related_name='receipt', to='core.transfer')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='TransferFailure',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified')),
                ('transfer', models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, related_name='failure', to='core.transfer')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='TransferCancellation',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified')),
                ('canceled_by', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, to='auth.user')),
                ('transfer', models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, related_name='cancellation', to='core.transfer')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='StoreRSAKeyPair',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('public_key_pem', models.TextField()),
                ('private_key_pem', models.TextField()),
                ('store', models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, related_name='rsa', to='core.store')),
            ],
        ),
        migrations.CreateModel(
            name='RaidenClientAccount',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('raiden', models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, related_name='raiden_account', to='raiden.raiden')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='PaymentConfirmation',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified')),
                ('payment', models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, related_name='confirmation', to='core.payment')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.AddField(
            model_name='payment',
            name='route',
            field=models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, to='core.paymentroute'),
        ),
        migrations.CreateModel(
            name='RaidenTransferExecution',
            fields=[
                ('transferexecution_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='core.transferexecution')),
                ('payment', models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, to='raiden.payment')),
            ],
            options={
                'abstract': False,
            },
            bases=('core.transferexecution',),
        ),
        migrations.CreateModel(
            name='RaidenPayment',
            fields=[
                ('payment_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='core.payment')),
                ('payment', models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, to='raiden.payment')),
            ],
            options={
                'abstract': False,
            },
            bases=('core.payment',),
        ),
        migrations.CreateModel(
            name='Debit',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified')),
                ('amount', hub20.apps.ethereum_money.fields.EthereumTokenAmountField(decimal_places=18, max_digits=32)),
                ('reference_id', models.PositiveIntegerField()),
                ('book', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='debits', to='core.book')),
                ('currency', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, to='ethereum_money.ethereumtoken')),
                ('reference_type', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='contenttypes.contenttype')),
            ],
            options={
                'abstract': False,
                'unique_together': {('book', 'reference_type', 'reference_id')},
            },
        ),
        migrations.CreateModel(
            name='Credit',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified')),
                ('amount', hub20.apps.ethereum_money.fields.EthereumTokenAmountField(decimal_places=18, max_digits=32)),
                ('reference_id', models.PositiveIntegerField()),
                ('book', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='credits', to='core.book')),
                ('currency', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, to='ethereum_money.ethereumtoken')),
                ('reference_type', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='contenttypes.contenttype')),
            ],
            options={
                'abstract': False,
                'unique_together': {('book', 'reference_type', 'reference_id')},
            },
        ),
        migrations.CreateModel(
            name='BlockchainTransferExecution',
            fields=[
                ('transferexecution_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='core.transferexecution')),
                ('transaction', models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, to='blockchain.transaction')),
            ],
            options={
                'abstract': False,
            },
            bases=('core.transferexecution',),
        ),
        migrations.CreateModel(
            name='BlockchainPaymentRoute',
            fields=[
                ('paymentroute_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='core.paymentroute')),
                ('payment_window', django.contrib.postgres.fields.ranges.IntegerRangeField()),
                ('account', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='blockchain_routes', to='blockchain.baseethereumaccount')),
                ('chain', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='blockchain.chain')),
            ],
            options={
                'abstract': False,
            },
            bases=('core.paymentroute',),
        ),
        migrations.CreateModel(
            name='BlockchainPayment',
            fields=[
                ('payment_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='core.payment')),
                ('transaction', models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, to='blockchain.transaction')),
            ],
            options={
                'abstract': False,
            },
            bases=('core.payment',),
        ),
        migrations.CreateModel(
            name='RaidenPaymentRoute',
            fields=[
                ('paymentroute_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='core.paymentroute')),
                ('payment_window', models.DurationField(default=hub20.apps.core.models.payments.calculate_raiden_payment_window)),
                ('identifier', models.BigIntegerField(default=hub20.apps.core.models.payments.generate_payment_order_id, unique=True)),
                ('raiden', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='payment_routes', to='raiden.raiden')),
            ],
            options={
                'unique_together': {('raiden', 'identifier')},
            },
            bases=('core.paymentroute',),
        ),
        migrations.CreateModel(
            name='Checkout',
            fields=[
                ('paymentorder_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='core.paymentorder')),
                ('external_identifier', models.TextField()),
                ('requester_ip', models.GenericIPAddressField(null=True)),
                ('store', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='core.store')),
            ],
            options={
                'abstract': False,
            },
            bases=('core.paymentorder',),
        ),
    ]



from typing import List, Optional

from pydantic import BaseModel

RPCProviderURL = str
FaucetURL = str


class NativeCurrency(BaseModel):
    name: str
    symbol: str
    decimals: int


class BlockchainExplorer(BaseModel):
    name: str
    url: str
    standard: str


class Chain(BaseModel):
    name: str
    chain: str
    network: Optional[str]
    icon: Optional[str]
    rpc: List[RPCProviderURL]
    faucets: List[FaucetURL]
    nativeCurrency: NativeCurrency
    infoURL: str
    shortName: str
    chainId: int
    networkId: int
    slip44: Optional[int]
    explorers: Optional[List[BlockchainExplorer]]



from django.apps import AppConfig


class BlockchainConfig(AppConfig):
    name = "hub20.apps.blockchain"

    def ready(self):
        from . import signals  # noqa
        from . import tasks  # noqa



import json
import os
import uuid
from datetime import timedelta
from decimal import Decimal

from celery import Celery
from celery.schedules import crontab
from django.conf import settings
from hexbytes import HexBytes
from kombu.serialization import register
from web3.datastructures import AttributeDict

NODE_HEALTH_CHECK_INTERVAL = 15


class Web3Encoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, AttributeDict):
            return {
                "__type__": "attrdict",
                "value": json.dumps({k: v for k, v in obj.items()}, cls=self.__class__),
            }
        elif isinstance(obj, HexBytes):
            return {"__type__": "hexbytes", "value": obj.hex()}
        elif isinstance(obj, uuid.UUID):
            return {"__type__": "uuid", "value": str(obj)}
        elif isinstance(obj, Decimal):
            return {"__type__": "decimal", "value": str(obj)}
        else:
            return json.JSONEncoder.default(self, obj)


def web3_decoder(obj):
    try:
        if obj["__type__"] == "attrdict":
            return AttributeDict(json.loads(obj["value"], object_hook=web3_decoder))
        elif obj["__type__"] == "hexbytes":
            return HexBytes(obj["value"])
        elif obj["__type__"] == "uuid":
            return uuid.UUID(obj["value"])
        elif obj["__type__"] == "decimal":
            return Decimal(obj["value"])
        else:
            return obj
    except KeyError:
        return obj


def web3_serializer(obj):
    return json.dumps(obj, cls=Web3Encoder)


def web3_deserializer(obj):
    return json.loads(obj, object_hook=web3_decoder)


register(
    "web3",
    web3_serializer,
    web3_deserializer,
    content_type="application/json",
    content_encoding="utf-8",
)


class Hub20CeleryConfig:
    name = "Hub20"

    broker_url = "memory" if "HUB20_TEST" in os.environ else settings.CELERY_BROKER_URL
    broker_use_ssl = "HUB20_BROKER_USE_SSL" in os.environ
    beat_schedule = {
        "clear-expired-sessions": {
            "task": "hub20.apps.core.tasks.clear_expired_sessions",
            "schedule": crontab(minute="*/30"),
        },
        "execute-transfers": {
            "task": "hub20.apps.core.tasks.execute_pending_transfers",
            "schedule": crontab(),
        },
        "check-providers-connected": {
            "task": "hub20.apps.blockchain.tasks.check_providers_are_connected",
            "schedule": timedelta(seconds=NODE_HEALTH_CHECK_INTERVAL),
        },
        "check-providers-synced": {
            "task": "hub20.apps.blockchain.tasks.check_providers_are_synced",
            "schedule": timedelta(seconds=NODE_HEALTH_CHECK_INTERVAL),
        },
        "check-chain-reorgs": {
            "task": "hub20.apps.blockchain.tasks.check_chains_were_reorganized",
            "schedule": timedelta(seconds=10),
        },
        "reset-inactive-providers": {
            "task": "hub20.apps.blockchain.tasks.reset_inactive_providers",
            "schedule": crontab(minute="*/5"),
        },
        "refresh-priority-fee-cache": {
            "task": "hub20.apps.blockchain.tasks.refresh_max_priority_fee",
            "schedule": timedelta(seconds=30),
        },
    }
    task_always_eager = "HUB20_TEST" in os.environ
    task_eager_propagates = "HUB20_TEST" in os.environ
    task_serializer = "web3"
    accept_content = ["web3", "json"]


app = Celery()
app.config_from_object(Hub20CeleryConfig)
app.autodiscover_tasks()



from django import forms

from . import models, validators

TOKEN_FILTER_QS = models.EthereumToken.tradeable.all()


class TokenlistStandardURLField(forms.URLField):
    default_validators = [validators.tokenlist_uri_validator]


class TokenForm(forms.ModelForm):
    logoURI = TokenlistStandardURLField(required=False)

    class Meta:
        model = models.EthereumToken
        fields = ("chain", "address", "symbol", "name", "decimals", "logoURI", "is_listed")


class TokenListForm(forms.ModelForm):
    tokens = forms.ModelMultipleChoiceField(queryset=TOKEN_FILTER_QS)

    class Meta:
        model = models.TokenList
        fields = ("name", "description", "tokens")


class WrappedTokenForm(forms.ModelForm):
    wrapped = forms.ModelChoiceField(queryset=TOKEN_FILTER_QS)
    wrapper = forms.ModelChoiceField(queryset=TOKEN_FILTER_QS)

    class Meta:
        model = models.WrappedToken
        fields = ("wrapped", "wrapper")


class StableTokenPairForm(forms.ModelForm):
    token = forms.ModelChoiceField(queryset=TOKEN_FILTER_QS)

    class Meta:
        model = models.StableTokenPair
        fields = ("token", "currency", "algorithmic_peg")



from rest_framework.routers import SimpleRouter

from . import views

app_name = "ethereum_money"


class TokenRouter(SimpleRouter):
    def get_lookup_regex(self, viewset, lookup_prefix=""):
        return f"{lookup_prefix}(?P<chain_id>\d+)-(?P<address>0x[0-9a-fA-F]{{40}})"


token_router = TokenRouter(trailing_slash=False)
token_router.register(r"tokens", views.TokenViewSet, basename="token")


router = SimpleRouter(trailing_slash=False)
router.register(r"tokenlists", views.TokenListViewSet, basename="tokenlist")

urlpatterns = token_router.urls + router.urls



from urllib.parse import urlparse

from django.core.exceptions import ValidationError
from ethereum.utils import check_checksum


def validate_checksumed_address(address):
    try:
        if not check_checksum(address):
            raise ValidationError(
                "%(address)s has an invalid checksum",
                params={"address": address},
            )
    except Exception:
        raise ValidationError(
            "%(address)s is not a valid ethereum address",
            params={"address": address},
        )


def uri_parsable_scheme_validator(schemes):
    def decorator(url):
        parsed = urlparse(url)

        errors = []

        if parsed.scheme not in schemes:
            errors.append(
                ValidationError(
                    "Scheme %(value)s is not acceptable",
                    code="required",
                    params=dict(value=parsed.scheme),
                )
            )

        if not parsed.netloc:
            errors.append(
                ValidationError(
                    "Could not find host/location for %(value)s ",
                    code="required",
                    params=dict(value=url),
                )
            )

        if errors:
            raise ValidationError(errors)

    return decorator


web3_url_validator = uri_parsable_scheme_validator(("http", "https", "ws", "wss", "ipc"))



import asyncio
import logging
from signal import SIGINT, SIGTERM

logger = logging.getLogger(__name__)


async def shutdown(signal, loop):
    logger.info(f"{signal.name} received. Going to shutdown...")
    for task in asyncio.all_tasks():
        if task is not asyncio.current_task():
            task.cancel()

    loop.stop()


def add_shutdown_handlers(loop):
    for signal in (SIGINT, SIGTERM):
        loop.add_signal_handler(signal, lambda: asyncio.create_task(shutdown(signal, loop)))



import logging
import uuid
from enum import Enum
from typing import Union

from asgiref.sync import async_to_sync
from channels.generic.websocket import JsonWebsocketConsumer

from . import models

logger = logging.getLogger(__name__)


class Events(Enum):
    INTERNAL_DEPOSIT_CONFIRMED = "internal.deposit.confirmed"
    BLOCKCHAIN_BLOCK_CREATED = "blockchain.block.created"
    BLOCKCHAIN_DEPOSIT_RECEIVED = "blockchain.deposit.received"
    BLOCKCHAIN_DEPOSIT_BROADCAST = "blockchain.deposit.broadcast"
    BLOCKCHAIN_DEPOSIT_CONFIRMED = "blockchain.deposit.confirmed"
    BLOCKCHAIN_ROUTE_EXPIRED = "blockchain.payment_route.expired"
    ETHEREUM_NODE_UNAVAILABLE = "ethereum_node.unavailable"
    ETHEREUM_NODE_OK = "ethereum_node.ok"
    RAIDEN_ROUTE_EXPIRED = "raiden.payment_route.expired"
    RAIDEN_DEPOSIT_RECEIVED = "raiden.deposit.received"
    RAIDEN_DEPOSIT_CONFIRMED = "raiden.deposit.confirmed"


def accept_subprotocol(consumer):
    try:
        subprotocol = consumer.scope["subprotocols"][0]
        consumer.accept(subprotocol)
    except IndexError:
        consumer.accept()


class SessionEventsConsumer(JsonWebsocketConsumer):
    @classmethod
    def get_group_name(cls, session_key: str) -> str:
        return f"events.{session_key}"

    def connect(self):
        session = self.scope["session"]
        session.save()
        accept_subprotocol(self)
        group_name = self.__class__.get_group_name(session.session_key)
        async_to_sync(self.channel_layer.group_add)(group_name, self.channel_name)
        logger.debug(f"Session Event consumer {group_name} connected")

    def notify_event(self, message):
        logger.info(f"Message received: {message}")
        message.pop("type", None)
        event = message.pop("event", "notification")
        logger.debug(f"Sending {event} notification... {message}")

        self.send_json({"event": event, "data": message})


class CheckoutConsumer(JsonWebsocketConsumer):
    @classmethod
    def get_group_name(cls, checkout_id: Union[uuid.UUID, str]) -> str:
        uid = uuid.UUID(str(checkout_id))
        return f"checkout.{uid.hex}"

    def connect(self):
        checkout_id = self.scope["url_route"]["kwargs"].get("pk")

        if not models.Checkout.objects.filter(id=checkout_id).first():
            self.close()

        group_name = self.__class__.get_group_name(checkout_id)
        async_to_sync(self.channel_layer.group_add)(group_name, self.channel_name)

        accept_subprotocol(self)
        logger.info(f"Checkout consumer {group_name} connected")

    def checkout_event(self, message):
        logger.info(f"Message received: {message}")
        message.pop("type", None)
        message["event"] = message.pop("event_name", None)

        logger.debug(f"Sending {message['event']} notification on {self.channel_name}...")
        self.send_json(message)



import factory

from hub20.apps.blockchain.factories import BaseWalletFactory

from .models import HierarchicalDeterministicWallet, KeystoreAccount


class KeystoreAccountFactory(BaseWalletFactory):
    class Meta:
        model = KeystoreAccount


class HDWalletFactory(BaseWalletFactory):
    index = factory.LazyFunction(lambda: HierarchicalDeterministicWallet.objects.count())

    class Meta:
        model = HierarchicalDeterministicWallet


EthereumAccountFactory = BaseWalletFactory

__all__ = ["EthereumAccountFactory", "HDWalletFactory", "KeystoreAccount"]



from unittest.mock import patch

import pytest
from django.test import TestCase

from hub20.apps.blockchain.tests.mocks import BlockMock, Web3Mock
from hub20.apps.core.factories import CheckoutFactory
from hub20.apps.ethereum_money.tasks import record_token_transfers
from hub20.apps.ethereum_money.tests.mocks import (
    Erc20LogFilterMock,
    Erc20TransferDataMock,
    Erc20TransferReceiptMock,
)


@pytest.mark.django_db(transaction=True)
class BaseTestCase(TestCase):
    pass


class PaymentTransferTestCase(BaseTestCase):
    def setUp(self):
        self.w3 = Web3Mock

    @patch("hub20.apps.ethereum_money.tasks.Web3Provider")
    @patch("hub20.apps.ethereum_money.tasks.make_web3")
    def test_can_detect_erc20_transfers(self, make_web3_mock, MockWeb3Provider):

        checkout = CheckoutFactory()
        route = checkout.routes.select_subclasses().first()

        self.assertIsNotNone(route)

        make_web3_mock

        transaction_params = dict(
            blockNumber=checkout.currency.chain.highest_block,
            recipient=route.account.address,
            amount=checkout.as_token_amount,
        )

        tx_data = Erc20TransferDataMock(**transaction_params)
        tx_receipt = Erc20TransferReceiptMock(
            hash=tx_data.hash,
            blockHash=tx_data.blockHash,
            from_address=tx_data["from"],
            transactionIndex=tx_data.transactionIndex,
            **transaction_params,
        )
        block_data = BlockMock(
            hash=tx_data.blockHash, number=tx_data.blockNumber, transactions=[tx_data.hash]
        )

        event_data = Erc20LogFilterMock(
            transactionHash=tx_data.hash,
            amount=checkout.as_token_amount,
            recipient=route.account.address,
        )

        make_web3_mock.return_value = self.w3
        self.w3.eth.get_transaction.return_value = tx_data
        self.w3.eth.get_transaction_receipt.return_value = tx_receipt
        self.w3.eth.get_block.return_value = block_data

        record_token_transfers(
            chain_id=self.w3.eth.chain_id,
            event_data=event_data,
            provider_url=self.w3.provider.endpoint_uri,
        )

        self.assertEqual(checkout.status, checkout.STATUS.paid)


__all__ = ["PaymentTransferTestCase"]



from django.contrib.auth import get_user_model
from django.db import models

from hub20.apps.ethereum_money.models import EthereumToken as Token

User = get_user_model()


class UserPreferences(models.Model):
    user = models.OneToOneField(User, related_name="preferences", on_delete=models.CASCADE)
    tokens = models.ManyToManyField(Token)


__all__ = ["UserPreferences"]



import asyncio
import logging

import celery_pubsub
from asgiref.sync import sync_to_async
from requests.exceptions import ConnectionError, HTTPError
from web3 import Web3
from web3.exceptions import TransactionNotFound

from hub20.apps.blockchain.app_settings import BLOCK_SCAN_RANGE
from hub20.apps.blockchain.exceptions import Web3UnsupportedMethod
from hub20.apps.blockchain.models import Chain, Web3Provider

from .web3 import make_web3

logger = logging.getLogger(__name__)
BLOCK_CREATION_INTERVAL = 10  # In seconds
PENDING_TX_POLLING_INTERVAL = 2  # In seconds

PENDING_TX_FILTER_REGISTRY = {}


def _get_pending_tx_filter(w3):
    global PENDING_TX_FILTER_REGISTRY
    try:
        return PENDING_TX_FILTER_REGISTRY[w3.eth.chain_id]
    except KeyError:
        try:
            eth_filter = w3.eth.filter("latest")
        except ValueError:
            raise Web3UnsupportedMethod("filter method not supported")

        PENDING_TX_FILTER_REGISTRY[w3.eth.chain_id] = eth_filter
        return eth_filter


async def generate_blocks(w3: Web3, chain: Chain):
    current_block = w3.eth.block_number
    start = chain.highest_block
    stop = min(current_block, chain.highest_block + BLOCK_SCAN_RANGE)
    for block_number in range(start, stop):
        logger.debug(f"Getting block #{block_number} from {chain}")
        yield w3.eth.get_block(block_number, full_transactions=True)


async def generate_tx_data(w3: Web3, transaction_list):
    for tx_data in transaction_list:
        try:
            tx_receipt = w3.eth.get_transaction_receipt(tx_data.hash)
            yield (tx_data, tx_receipt)
        except TransactionNotFound:
            pass


async def process_mined_blocks():
    while True:
        providers = await sync_to_async(list)(Web3Provider.available.select_related("chain"))
        for provider in providers:
            chain = provider.chain
            w3: Web3 = make_web3(provider=provider)
            logger.info(f"Getting blocks for {chain.name}")
            async for block_data in generate_blocks(w3=w3, chain=provider.chain):
                block_number = block_data.number
                logger.info(f"Processing block #{block_number} on {provider}")
                await sync_to_async(celery_pubsub.publish)(
                    "blockchain.mined.block",
                    chain_id=w3.eth.chain_id,
                    block_data=block_data,
                    provider_url=provider.url,
                )
                chain.highest_block = block_number
                logger.debug(f"Updating chain height to {block_number}")
            await sync_to_async(chain.save)()
        await asyncio.sleep(BLOCK_CREATION_INTERVAL)


async def process_pending_transactions():
    while True:
        providers = await sync_to_async(list)(
            Web3Provider.available.filter(supports_pending_filters=True).select_related("chain")
        )
        for provider in providers:
            w3: Web3 = make_web3(provider=provider)
            try:
                tx_filter = _get_pending_tx_filter(w3)
                for tx in tx_filter.get_new_entries():
                    tx_hash = tx.hex()
                    logger.info(f"Tx {tx_hash} broadcast by {provider.hostname}")
                    try:
                        tx_data = w3.eth.getTransaction(tx_hash)
                        logger.debug(f"{tx_data['to']} to {tx_data['from']}")
                        await sync_to_async(celery_pubsub.publish)(
                            "blockchain.broadcast.transaction",
                            chain_id=w3.eth.chain_id,
                            transaction_data=tx_data,
                        )
                    except TransactionNotFound:
                        logger.debug(f"Pending tx {tx_hash} not found on {provider.hostname}")
            except (Web3UnsupportedMethod, ValueError):
                logger.warning(
                    f"{provider.hostname} does not support pending filters and will be disabled"
                )
                await sync_to_async(Web3Provider.objects.filter(id=provider).update)(
                    supports_pending_filters=False
                )
            except (HTTPError, ConnectionError):
                logger.warning(f"Failed to connect to {provider.hostname}")
                await sync_to_async(celery_pubsub.publish)(
                    "node.connection.nok",
                    chain_id=provider.chain_id,
                    provider_url=provider.url,
                )
            except Exception:
                logger.exception(f"Failed to get pending txs from {provider.hostname}")

                # We remove the filter in case it was uninstalled by the server, so that we
                # can use a new one
                PENDING_TX_FILTER_REGISTRY.pop(w3.eth.chain_id, None)
        await asyncio.sleep(PENDING_TX_POLLING_INTERVAL)


__all__ = [
    "BLOCK_CREATION_INTERVAL",
    "process_mined_blocks",
    "process_pending_transactions",
]



import binascii
import logging

from django.core.management.base import BaseCommand
from web3.exceptions import TransactionNotFound

from hub20.apps.blockchain.client import make_web3
from hub20.apps.blockchain.models import Chain, Transaction, TransactionDataRecord

logger = logging.getLogger(__name__)


class Command(BaseCommand):
    help = "Records transaction data into database"

    def add_arguments(self, parser):
        parser.add_argument("transactions", metavar="N", nargs="+", type=str)
        parser.add_argument("--chain", dest="chain_id", required=True, type=int)

    def handle(self, *args, **options):

        chain_id = options["chain_id"]
        chain = Chain.active.get(id=chain_id)
        w3 = make_web3(provider=chain.provider)

        txs = options["transactions"]
        already_recorded = Transaction.objects.filter(
            block__chain_id=chain_id, hash__in=txs
        ).values_list("hash", flat=True)

        if already_recorded:
            logger.info(f"Transactions {', '.join(already_recorded)} already recorded")

        to_record = set(txs) - set(already_recorded)

        for tx_hash in to_record:
            try:
                tx_data = w3.eth.get_transaction(tx_hash)
                tx_receipt = w3.eth.get_transaction_receipt(tx_hash)
                block_data = w3.eth.get_block(tx_data.blockHash)
                TransactionDataRecord.make(chain_id=chain_id, tx_data=tx_data)
                Transaction.make(
                    chain_id=chain_id,
                    block_data=block_data,
                    tx_receipt=tx_receipt,
                )
            except binascii.Error:
                logger.info(f"{tx_hash} is not a valid transaction hash")
            except TransactionNotFound:
                logger.info(f"{tx_hash} not found")



from django.conf import settings

START_BLOCK_NUMBER = int(getattr(settings, "BLOCKCHAIN_START_BLOCK_NUMBER", 0) or 0)
BLOCK_SCAN_RANGE = int(getattr(settings, "BLOCKCHAIN_SCAN_BLOCK_RANGE", 0) or 5000)
FETCH_BLOCK_TASK_PRIORITY = int(getattr(settings, "BLOCKCHAIN_FETCH_BLOCK_PRIORITY", 0) or 9)



from rest_framework import permissions


class IsTokenListOwner(permissions.IsAuthenticated):
    def has_object_permission(self, request, view, obj):
        return request.user.is_authenticated and obj.user == request.user


class IsAdminOrReadOnly(permissions.IsAuthenticated):
    def has_permission(self, request, view):
        if request.method.upper() in permissions.SAFE_METHODS:
            return True

        return request.user.is_staff



from web3.exceptions import LogTopicError

EIP20_ABI = [
    {
        "constant": True,
        "inputs": [],
        "name": "name",
        "outputs": [{"name": "", "type": "string"}],
        "payable": False,
        "stateMutability": "view",
        "type": "function",
    },
    {
        "constant": False,
        "inputs": [{"name": "_spender", "type": "address"}, {"name": "_value", "type": "uint256"}],
        "name": "approve",
        "outputs": [{"name": "", "type": "bool"}],
        "payable": False,
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "constant": True,
        "inputs": [],
        "name": "totalSupply",
        "outputs": [{"name": "", "type": "uint256"}],
        "payable": False,
        "stateMutability": "view",
        "type": "function",
    },
    {
        "constant": False,
        "inputs": [
            {"name": "_from", "type": "address"},
            {"name": "_to", "type": "address"},
            {"name": "_value", "type": "uint256"},
        ],
        "name": "transferFrom",
        "outputs": [{"name": "", "type": "bool"}],
        "payable": False,
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "constant": True,
        "inputs": [],
        "name": "decimals",
        "outputs": [{"name": "", "type": "uint8"}],
        "payable": False,
        "stateMutability": "view",
        "type": "function",
    },
    {
        "constant": True,
        "inputs": [{"name": "_owner", "type": "address"}],
        "name": "balanceOf",
        "outputs": [{"name": "", "type": "uint256"}],
        "payable": False,
        "stateMutability": "view",
        "type": "function",
    },
    {
        "constant": True,
        "inputs": [],
        "name": "symbol",
        "outputs": [{"name": "", "type": "string"}],
        "payable": False,
        "stateMutability": "view",
        "type": "function",
    },
    {
        "constant": False,
        "inputs": [{"name": "_to", "type": "address"}, {"name": "_value", "type": "uint256"}],
        "name": "transfer",
        "outputs": [{"name": "", "type": "bool"}],
        "payable": False,
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "constant": True,
        "inputs": [{"name": "_owner", "type": "address"}, {"name": "_spender", "type": "address"}],
        "name": "allowance",
        "outputs": [{"name": "", "type": "uint256"}],
        "payable": False,
        "stateMutability": "view",
        "type": "function",
    },
    {
        "anonymous": False,
        "inputs": [
            {"indexed": True, "name": "_from", "type": "address"},
            {"indexed": True, "name": "_to", "type": "address"},
            {"indexed": False, "name": "_value", "type": "uint256"},
        ],
        "name": "Transfer",
        "type": "event",
    },
    {
        "anonymous": False,
        "inputs": [
            {"indexed": True, "name": "_owner", "type": "address"},
            {"indexed": True, "name": "_spender", "type": "address"},
            {"indexed": False, "name": "_value", "type": "uint256"},
        ],
        "name": "Approval",
        "type": "event",
    },
]


ERC223_ABI = [
    {
        "constant": True,
        "inputs": [],
        "name": "name",
        "outputs": [{"name": "", "type": "string"}],
        "payable": False,
        "stateMutability": "view",
        "type": "function",
    },
    {
        "constant": False,
        "inputs": [{"name": "_spender", "type": "address"}, {"name": "_value", "type": "uint256"}],
        "name": "approve",
        "outputs": [{"name": "success", "type": "bool"}],
        "payable": False,
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "constant": True,
        "inputs": [],
        "name": "totalSupply",
        "outputs": [{"name": "supply", "type": "uint256"}],
        "payable": False,
        "stateMutability": "view",
        "type": "function",
    },
    {
        "constant": True,
        "inputs": [],
        "name": "multiplier",
        "outputs": [{"name": "", "type": "uint256"}],
        "payable": False,
        "stateMutability": "view",
        "type": "function",
    },
    {
        "constant": False,
        "inputs": [
            {"name": "_from", "type": "address"},
            {"name": "_to", "type": "address"},
            {"name": "_value", "type": "uint256"},
        ],
        "name": "transferFrom",
        "outputs": [{"name": "success", "type": "bool"}],
        "payable": False,
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "constant": True,
        "inputs": [{"name": "", "type": "address"}],
        "name": "balances",
        "outputs": [{"name": "", "type": "uint256"}],
        "payable": False,
        "stateMutability": "view",
        "type": "function",
    },
    {
        "constant": True,
        "inputs": [],
        "name": "decimals",
        "outputs": [{"name": "decimals", "type": "uint8"}],
        "payable": False,
        "stateMutability": "view",
        "type": "function",
    },
    {
        "constant": True,
        "inputs": [],
        "name": "_decimals",
        "outputs": [{"name": "", "type": "uint8"}],
        "payable": False,
        "stateMutability": "view",
        "type": "function",
    },
    {
        "constant": False,
        "inputs": [],
        "name": "transferFunds",
        "outputs": [],
        "payable": False,
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "constant": True,
        "inputs": [],
        "name": "version",
        "outputs": [{"name": "", "type": "string"}],
        "payable": False,
        "stateMutability": "view",
        "type": "function",
    },
    {
        "constant": True,
        "inputs": [{"name": "_owner", "type": "address"}],
        "name": "balanceOf",
        "outputs": [{"name": "balance", "type": "uint256"}],
        "payable": False,
        "stateMutability": "view",
        "type": "function",
    },
    {
        "constant": True,
        "inputs": [],
        "name": "owner_address",
        "outputs": [{"name": "", "type": "address"}],
        "payable": False,
        "stateMutability": "view",
        "type": "function",
    },
    {
        "constant": True,
        "inputs": [],
        "name": "symbol",
        "outputs": [{"name": "", "type": "string"}],
        "payable": False,
        "stateMutability": "view",
        "type": "function",
    },
    {
        "constant": False,
        "inputs": [{"name": "num", "type": "uint256"}],
        "name": "mint",
        "outputs": [],
        "payable": False,
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "constant": False,
        "inputs": [{"name": "_to", "type": "address"}, {"name": "_value", "type": "uint256"}],
        "name": "transfer",
        "outputs": [{"name": "success", "type": "bool"}],
        "payable": False,
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "constant": False,
        "inputs": [{"name": "num", "type": "uint256"}, {"name": "target", "type": "address"}],
        "name": "mintFor",
        "outputs": [],
        "payable": False,
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "constant": True,
        "inputs": [{"name": "_owner", "type": "address"}, {"name": "_spender", "type": "address"}],
        "name": "allowance",
        "outputs": [{"name": "remaining", "type": "uint256"}],
        "payable": False,
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [
            {"name": "initial_supply", "type": "uint256"},
            {"name": "decimal_units", "type": "uint8"},
            {"name": "token_name", "type": "string"},
            {"name": "token_symbol", "type": "string"},
        ],
        "payable": False,
        "stateMutability": "nonpayable",
        "type": "constructor",
    },
    {
        "anonymous": False,
        "inputs": [
            {"indexed": True, "name": "_to", "type": "address"},
            {"indexed": True, "name": "_num", "type": "uint256"},
        ],
        "name": "Minted",
        "type": "event",
    },
    {
        "anonymous": False,
        "inputs": [
            {"indexed": True, "name": "_from", "type": "address"},
            {"indexed": True, "name": "_to", "type": "address"},
            {"indexed": False, "name": "_value", "type": "uint256"},
        ],
        "name": "Transfer",
        "type": "event",
    },
    {
        "anonymous": False,
        "inputs": [
            {"indexed": True, "name": "_owner", "type": "address"},
            {"indexed": True, "name": "_spender", "type": "address"},
            {"indexed": False, "name": "_value", "type": "uint256"},
        ],
        "name": "Approval",
        "type": "event",
    },
]


def get_event_abi(abi, event_name):
    events = [it for it in abi if it["type"] == "event"]
    try:
        return [e for e in events if e["name"] == event_name].pop()
    except (IndexError, LogTopicError):
        return None


TRANSFER_EVENT_ABI = get_event_abi(abi=EIP20_ABI, event_name="Transfer")
APPROVAL_EVENT_ABI = get_event_abi(abi=EIP20_ABI, event_name="Approval")
MINT_EVENT_ABI = get_event_abi(abi=ERC223_ABI, event_name="Minted")



from dj_rest_auth.serializers import UserDetailsSerializer
from rest_framework import serializers


class UserProfileSerializer(UserDetailsSerializer):
    has_admin_access = serializers.BooleanField(source="is_staff", read_only=True)

    class Meta:
        model = UserDetailsSerializer.Meta.model
        fields = UserDetailsSerializer.Meta.fields + ("has_admin_access",)
        read_only_fields = UserDetailsSerializer.Meta.read_only_fields + ("has_admin_access",)



import decimal
import os
import random

import ethereum
from faker.providers import BaseProvider
from hexbytes import HexBytes


class EthereumProvider(BaseProvider):
    def ethereum_address(self):
        private_key = os.urandom(32)
        address = ethereum.utils.privtoaddr(private_key)
        return ethereum.utils.checksum_encode(address)

    def hex64(self):
        return HexBytes(os.urandom(32)).hex()

    def uint256(self):
        return decimal.Decimal(random.randint(1, 2 ** 256))



from django.db.models import BooleanField, Case, Q, Value, When
from django.db.models.query import QuerySet
from django.http import Http404
from django.shortcuts import get_object_or_404
from django_filters import rest_framework as filters
from eth_utils import is_address
from rest_framework import status
from rest_framework.decorators import action
from rest_framework.filters import OrderingFilter, SearchFilter
from rest_framework.mixins import CreateModelMixin, ListModelMixin, RetrieveModelMixin
from rest_framework.permissions import IsAdminUser, IsAuthenticated
from rest_framework.response import Response
from rest_framework.viewsets import GenericViewSet, ModelViewSet

from hub20.apps.blockchain.models import Chain

from . import models, serializers, tasks


class TokenFilter(filters.FilterSet):
    chain_id = filters.ModelChoiceFilter(queryset=Chain.active.all())
    listed = filters.BooleanFilter(label="listed", method="filter_listed")
    native = filters.BooleanFilter(label="native", method="filter_native")
    stable_tokens = filters.BooleanFilter(label="stable", method="filter_stable_tokens")
    fiat = filters.CharFilter(label="fiat", method="filter_fiat")

    def token_search(self, queryset, name, value):
        q_name = Q(name__istartswith=value)
        q_symbol = Q(symbol__iexact=value)
        q_chain_name = Q(chain__name__icontains=value)
        return queryset.filter(q_name | q_symbol | q_chain_name)

    def filter_listed(self, queryset, name, value):
        return queryset.exclude(ethereum_money_tokenlist_tokenlists__isnull=value)

    def filter_native(self, queryset, name, value):
        filtered_qs = queryset.filter if value else queryset.exclude
        return filtered_qs(address=models.EthereumToken.NULL_ADDRESS)

    def filter_stable_tokens(self, queryset, name, value):
        return queryset.exclude(stable_pair__token__isnull=value)

    def filter_fiat(self, queryset, name, value):
        return queryset.filter(stable_pair__currency__iexact=value)

    class Meta:
        model = models.EthereumToken
        ordering_fields = ("symbol", "chain_id")
        fields = ("chain_id", "symbol", "address", "listed", "native", "stable_tokens", "fiat")


class BaseTokenViewSet(GenericViewSet, ListModelMixin, CreateModelMixin, RetrieveModelMixin):
    serializer_class = serializers.HyperlinkedEthereumTokenSerializer
    filterset_class = TokenFilter
    filter_backends = (
        OrderingFilter,
        SearchFilter,
        filters.DjangoFilterBackend,
    )
    page_size = 50
    search_fields = ("name", "=symbol", "chain__name")
    ordering_fields = ("symbol", "name", "chain_id")
    ordering = ("-is_native", "chain_id", "symbol")

    def get_queryset(self) -> QuerySet:
        return models.EthereumToken.tradeable.annotate(
            is_native=Case(
                When(address=models.EthereumToken.NULL_ADDRESS, then=Value(True)),
                default=Value(False),
                output_field=BooleanField(),
            )
        )

    def get_object(self):
        address = self.kwargs["address"]
        if not is_address(address):
            raise Http404(f"{address} is not a valid token address")

        return get_object_or_404(
            models.EthereumToken, chain_id=self.kwargs["chain_id"], address=address
        )


class TokenViewSet(BaseTokenViewSet):
    def get_serializer_class(self):
        if self.action == "info":
            return serializers.TokenInfoSerializer
        return self.serializer_class

    @action(detail=True)
    def info(self, request, **kwargs):
        """
        Returns extra information that the hub operator has provided about this token.
        """
        token = self.get_object()
        serializer = self.get_serializer(instance=token)
        return Response(serializer.data)


class TokenListViewSet(GenericViewSet, ListModelMixin, RetrieveModelMixin):
    """
    Token lists (https://tokenlists.org) is community-led effort to
    curate lists of ERC20 tokens. Hub Operators can create their own
    token lists, and its users can choose freely which ones to use.
    """

    permission_classes = (IsAuthenticated,)
    serializer_class = serializers.TokenListSerializer
    queryset = models.TokenList.objects.all()

    def get_serializer_class(self):
        if self.action == "_import":
            return serializers.TokenListImportSerializer

        return serializers.TokenListSerializer

    @action(
        detail=False,
        methods=["post"],
        url_path="import",
        name="Import from URL",
        permission_classes=(IsAdminUser,),
    )
    def _import(self, request, **kwargs):
        """
        Fetches, validates and imports the token list from the
        provided URL. This process is executed on the background
        """
        serializer = self.get_serializer(data=request.data)

        if serializer.is_valid():
            tasks.import_token_list.delay(
                url=serializer.validated_data["url"],
                description=serializer.validated_data["description"],
            )
            return Response(serializer.data)
        else:
            return Response(serializer.errors, status=status.HTTP_400_BAD_REQUEST)


class UserTokenListViewSet(ModelViewSet):

    permission_classes = (IsAuthenticated,)
    serializer_class = serializers.UserTokenListSerializer

    def get_queryset(self):
        return self.request.user.token_lists.all()

    def get_serializer_class(self):
        if self.action == "clone":
            return serializers.UserTokenListCloneSerializer
        return serializers.UserTokenListSerializer

    @action(detail=False, methods=["post"])
    def clone(self, request, **kwargs):
        """
        Clone a token list from the site's base, allowing users to
        create their own token lists
        """
        serializer = self.get_serializer(data=request.data)

        if serializer.is_valid():
            serializer.save()
            headers = self.get_success_headers(serializer.data)
            return Response(serializer.data, status=status.HTTP_201_CREATED, headers=headers)
        else:
            return Response(serializer.errors, status=status.HTTP_400_BAD_REQUEST)



import logging

from django.contrib.auth import get_user_model
from django.contrib.contenttypes.models import ContentType
from django.core.management.base import BaseCommand
from django.db.models import QuerySet
from web3 import Web3

from hub20.apps.blockchain.app_settings import BLOCK_SCAN_RANGE
from hub20.apps.blockchain.client import make_web3
from hub20.apps.blockchain.models import BaseEthereumAccount, Chain, Transaction, Web3Provider
from hub20.apps.core.models.accounting import (
    ExternalAddressAccount,
    RaidenClientAccount,
    Treasury,
    UserAccount,
    WalletAccount,
)
from hub20.apps.ethereum_money.client import (
    index_account_erc20_approvals,
    index_account_erc20_transfers,
    index_account_erc223_transactions,
)
from hub20.apps.ethereum_money.models import EthereumToken
from hub20.apps.raiden.client.blockchain import (
    get_all_channel_deposits,
    get_all_service_deposits,
    index_all_token_network_events,
)
from hub20.apps.raiden.models import Payment as RaidenPayment, Raiden

User = get_user_model()
logger = logging.getLogger(__name__)


def index_token_events(w3: Web3, chain: Chain, accounts: QuerySet, tokens: QuerySet):
    starting_block = 0

    while starting_block < chain.highest_block:
        for token in tokens:
            chain = token.chain
            for account in accounts:
                end_block = min(starting_block + BLOCK_SCAN_RANGE, chain.highest_block)
                index_account_erc20_approvals(
                    w3=w3,
                    account=account,
                    token=token,
                    starting_block=starting_block,
                    end_block=end_block,
                )
                index_account_erc20_transfers(
                    w3=w3,
                    account=account,
                    token=token,
                    starting_block=starting_block,
                    end_block=end_block,
                )
                index_account_erc223_transactions(
                    w3=w3,
                    account=account,
                    token=token,
                    starting_block=starting_block,
                    end_block=end_block,
                )
        starting_block += BLOCK_SCAN_RANGE


class Command(BaseCommand):
    help = "Sets up accounting books and reconciles transactions and raiden payments"

    def handle(self, *args, **options):
        accounts = BaseEthereumAccount.objects.all()
        transaction_type = ContentType.objects.get_for_model(Transaction)

        for user in User.objects.all():
            UserAccount.objects.get_or_create(user=user)

        for wallet in accounts:
            WalletAccount.objects.get_or_create(account=wallet)

        # Index Transactions
        for provider in Web3Provider.available.all():
            chain = provider.chain
            w3 = make_web3(provider=provider)
            Treasury.objects.get_or_create(chain=chain)

            native_token = EthereumToken.make_native(chain=chain)
            tokens = EthereumToken.ERC20tokens.filter(chain=chain)

            index_token_events(w3=w3, chain=chain, accounts=accounts, tokens=tokens)
            index_all_token_network_events(w3=w3)

            for raiden in Raiden.objects.all():
                RaidenClientAccount.objects.get_or_create(raiden=raiden)
                get_all_service_deposits(w3=w3, raiden=raiden)
                get_all_channel_deposits(w3=w3, raiden=raiden)

            # Native Token Value Transfers
            for account in accounts:
                wallet_book = account.onchain_account.get_book(token=native_token)

                # Ethereum Transactions Received
                for tx in account.transactions.filter(to_address=account.address, value__gt=0):
                    amount = native_token.from_wei(tx.value)
                    params = dict(
                        reference_type=transaction_type,
                        reference_id=tx.id,
                        currency=native_token,
                        amount=amount.amount,
                    )
                    external_address_account, _ = ExternalAddressAccount.objects.get_or_create(
                        address=tx.from_address
                    )
                    external_address_book = external_address_account.get_book(token=native_token)

                    external_address_book.debits.get_or_create(**params)
                    wallet_book.credits.get_or_create(**params)

                # Ethereum Transactions Sent
                for tx in account.transactions.filter(from_address=account.address, value__gt=0):
                    amount = native_token.from_wei(tx.value)
                    params = dict(
                        reference_type=transaction_type,
                        reference_id=tx.id,
                        currency=native_token,
                        amount=amount.amount,
                    )
                    external_address_account, _ = ExternalAddressAccount.objects.get_or_create(
                        address=tx.to_address
                    )

                    external_address_book = external_address_account.get_book(token=ETH)

                    external_address_book.credits.get_or_create(**params)
                    wallet_book.debits.get_or_create(**params)

        # Raiden payments
        for raiden in Raiden.objects.all():
            payment_type = ContentType.objects.get_for_model(RaidenPayment)
            for channel in raiden.channels.all():
                logger.info(f"Recording entries for {channel}")
                for payment in channel.payments.all():
                    logger.info(f"Recording entries for {payment}")
                    params = dict(
                        reference_type=payment_type,
                        reference_id=payment.id,
                        amount=payment.amount,
                        currency=payment.token,
                    )

                    external_address_account, _ = ExternalAddressAccount.objects.get_or_create(
                        address=payment.partner_address
                    )

                    external_account_book = external_address_account.get_book(token=payment.token)
                    raiden_book = raiden.raiden_account.get_book(token=payment.token)

                    if payment.is_outgoing:
                        raiden_book.debits.get_or_create(**params)
                        external_account_book.credits.get_or_create(**params)
                    else:
                        external_account_book.debits.get_or_create(**params)
                        raiden_book.credits.get_or_create(**params)



import logging

import celery_pubsub
from celery import shared_task
from django.db.transaction import atomic
from requests.exceptions import ConnectionError

from .analytics import MAX_PRIORITY_FEE_TRACKER, get_historical_block_data
from .client import inspect_web3, make_web3
from .models import BaseEthereumAccount, Block, Transaction, Web3Provider
from .signals import block_sealed

logger = logging.getLogger(__name__)

# Tasks that are meant to be run periodically (no arguments)


@shared_task
def reset_inactive_providers():
    Web3Provider.objects.filter(is_active=False).update(synced=False, connected=False)


@shared_task
def refresh_max_priority_fee():
    for provider in Web3Provider.available.filter(supports_eip1559=True):
        try:
            w3 = make_web3(provider=provider)
            MAX_PRIORITY_FEE_TRACKER.set(w3.eth.chain_id, w3.eth.max_priority_fee)
        except Exception as exc:
            logger.info(f"Failed to get max priority fee from {provider.hostname}: {exc}")


@shared_task
def check_providers_configuration():
    for provider in Web3Provider.active.all():
        w3 = make_web3(provider=provider)
        configuration = inspect_web3(w3=w3)
        Web3Provider.objects.filter(id=provider.id).update(**configuration.dict())


@shared_task
def check_providers_are_connected():
    for provider in Web3Provider.active.all():
        logger.info(f"Checking status from {provider.hostname}")
        try:
            w3 = make_web3(provider=provider)
            is_online = w3.isConnected() and (w3.net.peer_count > 0)
        except ConnectionError:
            is_online = False
        except ValueError:
            # The node does not support the peer count method. Assume healthy.
            is_online = w3.isConnected()

        if provider.connected and not is_online:
            logger.info(f"Node {provider.hostname} went offline")
            celery_pubsub.publish(
                "node.connection.nok", chain_id=provider.chain_id, provider_url=provider.url
            )

        elif is_online and not provider.connected:
            logger.info(f"Node {provider.hostname} is back online")
            celery_pubsub.publish(
                "node.connection.ok", chain_id=provider.chain_id, provider_url=provider.url
            )


@shared_task
def check_providers_are_synced():
    for provider in Web3Provider.active.all():
        try:
            w3 = make_web3(provider=provider)
            is_synced = bool(not w3.eth.syncing)
        except ValueError:
            # The node does not support the eth_syncing method. Assume healthy.
            is_synced = True
        except ConnectionError:
            continue

        if provider.synced and not is_synced:
            celery_pubsub.publish(
                "node.sync.nok", chain_id=provider.chain_id, provider_url=provider.url
            )
        elif is_synced and not provider.synced:
            logger.info(f"Node {provider.hostname} is back in sync")
            celery_pubsub.publish(
                "node.sync.ok", chain_id=provider.chain_id, provider_url=provider.url
            )


@shared_task
def check_chains_were_reorganized():
    for provider in Web3Provider.active.all():
        with atomic():
            chain = provider.chain
            w3 = make_web3(provider=provider)
            block_number = w3.eth.block_number

            if chain.highest_block > block_number:
                chain.blocks.filter(number__gt=block_number).delete()

            chain.highest_block = block_number
            chain.save()


# Tasks that are setup to subscribe and handle events generated by the event streams
@shared_task
def save_historical_data(chain_id, block_data, provider_url):
    logger.debug(f"Adding {block_data} to {chain_id} historical data")
    block_history = get_historical_block_data(chain_id)
    block_history.push(block_data)


@shared_task
def notify_new_block(chain_id, block_data, provider_url):
    block_sealed.send(sender=Block, chain_id=chain_id, block_data=block_data)


@shared_task
def record_account_transactions(chain_id, block_data, provider_url):

    addresses = BaseEthereumAccount.objects.values_list("address", flat=True)

    txs = [
        t for t in block_data["transactions"] if (t["from"] in addresses or t["to"] in addresses)
    ]

    if len(txs) > 0:
        provider = Web3Provider.objects.get(url=provider_url)
        w3 = make_web3(provider=provider)
        assert chain_id == w3.eth.chain_id, f"{provider.hostname} not on chain #{chain_id}"

        for tx_data in txs:
            transaction_receipt = w3.eth.get_transaction_receipt(tx_data.hash)
            Transaction.make(
                chain_id=chain_id,
                tx_receipt=transaction_receipt,
                block_data=block_data,
            )


@shared_task
def set_node_connection_ok(chain_id, provider_url):
    logger.info(f"Setting node {provider_url} to online")
    Web3Provider.objects.filter(chain_id=chain_id, url=provider_url).update(connected=True)


@shared_task
def set_node_connection_nok(chain_id, provider_url):
    logger.info(f"Setting node {provider_url} to offline")
    Web3Provider.objects.filter(chain_id=chain_id, url=provider_url).update(connected=False)


@shared_task
def set_node_sync_ok(chain_id, provider_url):
    logger.info(f"Setting node {provider_url} to sync")
    Web3Provider.objects.filter(chain_id=chain_id, url=provider_url).update(synced=True)


@shared_task
def set_node_sync_nok(chain_id, provider_url):
    logger.info(f"Setting node {provider_url} to out-of-sync")
    Web3Provider.objects.filter(chain_id=chain_id, url=provider_url).update(synced=False)


celery_pubsub.subscribe("blockchain.mined.block", save_historical_data)
celery_pubsub.subscribe("blockchain.mined.block", notify_new_block)
celery_pubsub.subscribe("blockchain.mined.block", record_account_transactions)
celery_pubsub.subscribe("node.connection.ok", set_node_connection_ok)
celery_pubsub.subscribe("node.connection.nok", set_node_connection_nok)
celery_pubsub.subscribe("node.sync.ok", set_node_sync_ok)
celery_pubsub.subscribe("node.sync.nok", set_node_sync_nok)



from rest_framework.routers import SimpleRouter

from . import views

app_name = "blockchain"


router = SimpleRouter(trailing_slash=False)
router.register("", views.ChainViewSet, basename="chain")

urlpatterns = router.urls



from django.test import TestCase
from django.urls import reverse
from rest_framework.test import APIClient

from hub20.apps.ethereum_money import factories


class TokenViewSetTestCase(TestCase):
    def setUp(self):
        self.client = APIClient()
        self.token_list_url = reverse("token-list")

    def _get_token_url(self, token):
        return reverse(
            "token-detail",
            kwargs={"address": token.address, "chain_id": token.chain_id},
        )

    def _get_token_info_url(self, token):
        return f"{self._get_token_url(token)}/info"

    def test_anonymous_user_can_see_token_list(self):
        factories.Erc20TokenFactory()
        factories.ETHFactory()

        response = self.client.get(self.token_list_url)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(response.data), 2)

    def test_can_find_by_name_first_letters(self):
        factories.Erc20TokenFactory(name="Example Search")
        success_response = self.client.get(self.token_list_url, dict(search="exam"))
        self.assertEqual(len(success_response.data), 1)

        bad_response = self.client.get(self.token_list_url, dict(search="xyz"))
        self.assertEqual(len(bad_response.data), 0)

    def test_can_find_by_symbol(self):
        factories.Erc20TokenFactory(symbol="GOOD")
        success_response = self.client.get(self.token_list_url, dict(search="gOOd"))
        self.assertEqual(len(success_response.data), 1)

        bad_response = self.client.get(self.token_list_url, dict(search="BAD"))
        self.assertEqual(len(bad_response.data), 0)

    def test_can_filter_native_tokens(self):
        factories.ETHFactory()
        response = self.client.get(self.token_list_url, dict(native=True))
        self.assertEqual(len(response.data), 1)

    def test_can_filter_stable_tokens(self):
        factories.StableTokenFactory()
        response = self.client.get(self.token_list_url, dict(stable=True))
        self.assertEqual(len(response.data), 1)

    def test_can_filter_by_fiat_code(self):
        factories.StableTokenFactory(currency="EUR")

        good_response = self.client.get(self.token_list_url, dict(fiat="EUR"))
        self.assertEqual(len(good_response.data), 1)

        bad_response = self.client.get(self.token_list_url, dict(fiat="USD"))
        self.assertEqual(len(bad_response.data), 0)

    def test_anonymous_user_can_see_token(self):
        token = factories.Erc20TokenFactory()
        token_url = self._get_token_url(token)
        response = self.client.get(token_url)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.data["address"], str(token.address))

    def test_can_see_token_info(self):
        stable_pair = factories.StableTokenFactory(currency="INR")
        token = stable_pair.token
        response = self.client.get(self._get_token_info_url(token))
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.data["is_stable"], True)
        self.assertEqual(response.data["tracks_currency"], "INR")


__all__ = ["TokenViewSetTestCase"]



import logging

import requests
from django.core.management.base import BaseCommand

from hub20.apps.blockchain import models
from hub20.apps.blockchain.schemas import chainlist

logger = logging.getLogger(__name__)


class Command(BaseCommand):
    help = "Loads all chain information according to chainlist schema"

    def add_arguments(self, parser):
        parser.add_argument(
            "--url",
            dest="url",
            default="https://chainid.network/chains.json",
        )

    def handle(self, *args, **options):

        response = requests.get(options["url"])
        response.raise_for_status()

        for entry in response.json():
            try:
                chain_data = chainlist.Chain(**entry)
                chain, _ = models.Chain.objects.get_or_create(
                    id=chain_data.chainId,
                    defaults=dict(
                        name=chain_data.name,
                        is_mainnet=(chain_data.network == "mainnet"),
                        highest_block=0,
                    ),
                )
                native_token_data = chain_data.nativeCurrency
                models.NativeToken.objects.update_or_create(
                    chain=chain, defaults=native_token_data.dict()
                )

                for provider_url in chain_data.rpc:
                    models.Web3Provider.objects.get_or_create(
                        chain=chain,
                        url=provider_url,
                        defaults=dict(is_active=False),
                    )

                blockchain_explorers = chain_data.explorers or []
                for explorer in blockchain_explorers:
                    models.Explorer.objects.get_or_create(chain=chain, defaults=explorer.dict())
            except Exception as exc:
                logger.exception(exc)



import logging

from channels.auth import AuthMiddlewareStack
from django.contrib.auth.models import AnonymousUser
from django.db import close_old_connections
from rest_framework.authtoken.models import Token

logger = logging.getLogger(__name__)


class TokenAuthMiddleware:
    """
    Token authorization middleware for websocket consumers (django channels)
    """

    def __init__(self, inner):
        self.inner = inner

    async def __call__(self, scope, receive, send):
        headers = dict(scope["headers"])
        authorization_header = "authorization".encode()
        if authorization_header in headers:
            token_name, token_key = headers[authorization_header].decode().split()
            token = Token.objects.filter(key=token_key).first()
            if token_name.lower() == "token" and token is not None:
                scope["user"] = token.user
                close_old_connections()
            else:
                scope["user"] = AnonymousUser()
        return await self.inner(scope, receive, send)


TokenAuthMiddlewareStack = lambda inner: TokenAuthMiddleware(AuthMiddlewareStack(inner))



from __future__ import annotations

import asyncio
import logging
from datetime import datetime
from types import FunctionType
from typing import Any, Dict, List, Optional, Union

import requests
from asgiref.sync import sync_to_async
from django.contrib.auth import get_user_model
from django.utils.timezone import make_aware
from ethereum.utils import checksum_encode
from web3.datastructures import AttributeDict

from hub20.apps.blockchain.typing import Address
from hub20.apps.ethereum_money.models import EthereumTokenAmount
from hub20.apps.raiden.exceptions import RaidenConnectionError, RaidenPaymentError
from hub20.apps.raiden.models import Channel, Payment, Raiden, TokenNetwork

User = get_user_model()

logger = logging.getLogger(__name__)


def _make_request(url: str, method: str = "GET", **payload: Any) -> Union[List, Dict]:
    try:
        action = {
            "GET": requests.get,
            "PATCH": requests.patch,
            "PUT": requests.put,
            "POST": requests.post,
            "DELETE": requests.delete,
        }[method.strip().upper()]
    except KeyError:
        raise ValueError(f"{method} is not valid")

    try:
        assert isinstance(action, FunctionType)
        response = action(url, json=payload)
        response.raise_for_status()
        return response.json()
    except requests.exceptions.ConnectionError:
        raise RaidenConnectionError(f"Could not connect to {url}")


class RaidenClient:
    URL_BASE_PATH = "/api/v1"

    def __init__(self, raiden_account: Raiden) -> None:
        self.raiden = raiden_account

    def _parse_payment(self, payment_data: Dict, channel: Channel) -> Optional[AttributeDict]:
        event_name = payment_data.pop("event")
        payment_data.pop("token_address", None)

        if event_name == "EventPaymentReceivedSuccess":
            payment_data["sender_address"] = payment_data.pop("initiator")
            payment_data["receiver_address"] = self.raiden.address
        elif event_name == "EventPaymentSentSuccess":
            payment_data["sender_address"] = self.raiden.address
            payment_data["receiver_address"] = payment_data.pop("target")
        else:
            return None

        iso_time = payment_data.pop("log_time")

        payment_data["amount"] = channel.token.from_wei(payment_data.pop("amount")).amount
        payment_data["timestamp"] = make_aware(datetime.fromisoformat(iso_time))
        return AttributeDict(payment_data)

    def _refresh_channel(self, channel: Channel) -> Channel:
        channel_data = _make_request(self.channel_endpoint(channel=channel))
        assert isinstance(channel_data, dict)
        Channel.make(channel.raiden, **channel_data)
        channel.refresh_from_db()
        return channel

    @property
    def raiden_root_endpoint(self) -> str:
        return f"{self.raiden.url}{self.URL_BASE_PATH}"

    @property
    def raiden_channel_list_endpoint(self) -> str:
        return f"{self.raiden_root_endpoint}/channels"

    @property
    def raiden_token_list_endpoint(self) -> str:
        return f"{self.raiden_root_endpoint}/tokens"

    def channel_endpoint(self, channel: Channel) -> str:
        raiden_endpoint = self.raiden_root_endpoint
        return f"{raiden_endpoint}/channels/{channel.token.address}/{channel.partner_address}"

    def channel_payment_list_endpoint(self, channel: Channel) -> str:
        raiden_endpoint = self.raiden_root_endpoint
        return f"{raiden_endpoint}/payments/{channel.token.address}/{channel.partner_address}"

    def token_network_endpoint(self, token_network: TokenNetwork) -> str:
        raiden_endpoint = self.raiden_root_endpoint
        return f"{raiden_endpoint}/connections/{token_network.token.address}"

    def get_channels(self):
        return [
            Channel.make(self.raiden, **channel_data)
            for channel_data in _make_request(self.raiden_channel_list_endpoint)
        ]

    def get_new_payments(self):
        for channel in self.raiden.channels.all():
            offset = channel.payments.count()
            events = _make_request(
                self.channel_payment_list_endpoint(channel=channel), offset=offset
            )
            assert type(events) is list

            payments = [self._parse_payment(ev, channel) for ev in events]

            for payment_data in [payment for payment in payments if payment is not None]:
                Payment.make(channel, **payment_data)

    def get_token_addresses(self):
        return _make_request(self.raiden_token_list_endpoint)

    def get_status(self):
        try:
            response = _make_request(f"{self.raiden_root_endpoint}/status")
            assert type(response) is dict
            return response.get("status")
        except RaidenConnectionError:
            return "offline"

    def join_token_network(self, token_network: TokenNetwork, amount: EthereumTokenAmount):
        url = self.token_network_endpoint(token_network=token_network)
        return _make_request(url, method="PUT", funds=amount.as_wei)

    def leave_token_network(self, token_network: TokenNetwork):
        url = self.token_network_endpoint(token_network=token_network)
        return _make_request(url, method="DELETE")

    def make_channel_deposit(self, channel: Channel, amount: EthereumTokenAmount):
        channel = self._refresh_channel(channel)
        new_deposit = channel.deposit_amount + amount

        return _make_request(
            self.channel_endpoint(channel), method="PATCH", total_deposit=new_deposit.as_wei
        )

    def make_channel_withdraw(self, channel: Channel, amount: EthereumTokenAmount):
        channel = self._refresh_channel(channel)
        new_withdraw = channel.withdraw_amount + amount
        return _make_request(
            self.channel_endpoint(channel=channel),
            method="PATCH",
            total_withdraw=new_withdraw.as_wei,
        )

    def _ensure_valid_identifier(self, identifier: Optional[str] = None) -> Optional[int]:
        if not identifier:
            return None

        try:
            numeric_identifier = int(identifier)
        except ValueError:
            numeric_identifier = int(identifier.encode().hex(), 16)

        return numeric_identifier % Payment.MAX_IDENTIFIER_ID

    def transfer(
        self, amount: EthereumTokenAmount, address: Address, identifier: Optional[int] = None, **kw
    ) -> Optional[str]:
        url = f"{self.raiden_root_endpoint}/payments/{amount.currency.address}/{str(address)}"

        payload = dict(amount=amount.as_wei)

        if identifier:
            payload["identifier"] = identifier

        try:
            payment_data = _make_request(url, method="POST", **payload)
            assert isinstance(payment_data, dict)
            return payment_data.get("identifier")
        except requests.exceptions.HTTPError as error:
            logger.exception(error)

            error_code = error.response.status_code
            message = error.response.json().get("errors")
            raise RaidenPaymentError(error_code=error_code, message=message) from error

    @classmethod
    def get_node_account_address(cls, url) -> Address:
        response = _make_request(f"{url}{cls.URL_BASE_PATH}/address")
        assert isinstance(response, dict)
        return checksum_encode(response.get("our_address"))

    @classmethod
    def make(cls, url) -> RaidenClient:
        account_address: Address = cls.get_node_account_address(url)
        raiden_account, _ = Raiden.objects.get_or_create(address=account_address, url=url)
        return cls(raiden_account=raiden_account)

    @classmethod
    def select_for_transfer(
        cls,
        amount: EthereumTokenAmount,
        receiver: Optional[User] = None,
        address: Optional[Address] = None,
    ) -> Optional[RaidenClient]:
        if address is None:
            return None

        # Token is not part of a token network.
        if not hasattr(amount.currency, "tokennetwork"):
            return None

        token_channels = Channel.available.filter(
            token_network__token=amount.currency, balance__gte=amount.amount
        )

        if not token_channels.exists():
            return None

        if not amount.currency.tokennetwork.can_reach(address):
            return None

        raiden_account = Raiden.objects.first()

        return raiden_account and cls(raiden_account=raiden_account)


def raiden_periodic_response_handler(period=2):
    def decorator(handler):
        async def wrapper(*args, **kw):
            raiden_accounts = await sync_to_async(list)(Raiden.objects.exclude(web3_provider=None))
            raiden_clients = [RaidenClient(raiden_account=raiden) for raiden in raiden_accounts]

            while True:
                for raiden_client in raiden_clients:
                    try:
                        logger.debug(f"Running {handler.__name__} for {raiden_client.raiden.url}")
                        await sync_to_async(handler)(raiden_client=raiden_client, *args, **kw)
                    except RaidenConnectionError as exc:
                        logger.error(f"Failed to connect to raiden node: {exc}")
                    except Exception as exc:
                        logger.exception(f"Error on {handler.__name__}: {exc}")

                await asyncio.sleep(period)

        return wrapper

    return decorator


@raiden_periodic_response_handler(period=60)
def sync_channels(raiden_client: RaidenClient, **kw):
    raiden_client.get_channels()


@raiden_periodic_response_handler()
def sync_payments(raiden_client: RaidenClient, **kw):
    raiden_client.get_new_payments()



import logging
from typing import Optional, Tuple
from urllib.parse import urlparse

from pydantic import BaseModel
from web3 import Web3
from web3.exceptions import ExtraDataLengthError
from web3.middleware import geth_poa_middleware
from web3.providers import HTTPProvider, IPCProvider, WebsocketProvider
from web3.types import TxReceipt

from hub20.apps.blockchain import analytics
from hub20.apps.blockchain.exceptions import Web3TransactionError
from hub20.apps.blockchain.models import Web3Provider

logger = logging.getLogger(__name__)


class Web3ProviderConfiguration(BaseModel):
    client_version: Optional[str]
    supports_eip1559: bool
    supports_pending_filters: bool
    requires_geth_poa_middleware: bool


def eip1559_price_strategy(w3: Web3, *args, **kw):
    try:
        current_block = w3.eth.get_block("latest")
        return analytics.recommended_eip1559_gas_price(
            current_block, max_priority_fee=w3.eth.max_priority_fee
        )
    except Exception as exc:
        chain_id = w3.eth.chain_id
        logger.exception(f"Error when getting price estimate for {chain_id}", exc_info=exc)
        return analytics.estimate_gas_price(chain_id)


def historical_trend_price_strategy(w3: Web3, *args, **kw):
    return analytics.estimate_gas_price(w3.eth.chain_id)


def get_web3(provider_url: str) -> Web3:
    endpoint = urlparse(provider_url)

    provider_class = {
        "http": HTTPProvider,
        "https": HTTPProvider,
        "ws": WebsocketProvider,
        "wss": WebsocketProvider,
    }.get(endpoint.scheme, IPCProvider)

    w3 = Web3(provider_class(provider_url))
    return w3


def make_web3(provider: Web3Provider) -> Web3:
    w3 = get_web3(provider_url=provider.url)

    if provider.requires_geth_poa_middleware:
        w3.middleware_onion.inject(geth_poa_middleware, layer=0)

    price_strategy = (
        eip1559_price_strategy if provider.supports_eip1559 else historical_trend_price_strategy
    )
    w3.eth.setGasPriceStrategy(price_strategy)

    return w3


def inspect_web3(w3: Web3) -> Web3ProviderConfiguration:
    try:
        version: Optional[str] = w3.clientVersion
    except ValueError:
        version = None

    try:
        max_fee = w3.eth.max_priority_fee
        eip1559 = bool(type(max_fee) is int)
    except ValueError:
        eip1559 = False

    try:
        w3.eth.filter("pending")
        pending_filters = True
    except ValueError:
        pending_filters = False

    try:
        w3.eth.get_block("latest")
        requires_geth_poa_middleware = False
    except ExtraDataLengthError:
        requires_geth_poa_middleware = True

    return Web3ProviderConfiguration(
        client_version=version,
        supports_eip1559=eip1559,
        supports_pending_filters=pending_filters,
        requires_geth_poa_middleware=requires_geth_poa_middleware,
    )


def send_transaction(
    w3: Web3,
    contract_function,
    account,
    gas,
    contract_args: Optional[Tuple] = None,
    **kw,
) -> TxReceipt:
    nonce = kw.pop("nonce", w3.eth.getTransactionCount(account.address))

    transaction_params = {
        "chainId": int(w3.net.version),
        "nonce": nonce,
        "gasPrice": kw.pop("gas_price", w3.eth.generateGasPrice()),
        "gas": gas,
    }

    transaction_params.update(**kw)

    try:
        result = contract_function(*contract_args) if contract_args else contract_function()
        transaction_data = result.buildTransaction(transaction_params)
        signed = w3.eth.account.signTransaction(transaction_data, account.private_key)
        tx_hash = w3.eth.sendRawTransaction(signed.rawTransaction)
        return w3.eth.waitForTransactionReceipt(tx_hash)
    except ValueError as exc:
        try:
            if exc.args[0].get("message") == "nonce too low":
                logger.warning("Node reported that nonce is too low. Trying tx again...")
                kw["nonce"] = nonce + 1
                return send_transaction(
                    w3,
                    contract_function,
                    account,
                    gas,
                    contract_args=contract_args,
                    **kw,
                )
        except (AttributeError, IndexError):
            pass

        raise Web3TransactionError from exc



import logging

from django.db.models.signals import post_save
from django.dispatch import receiver

from hub20.apps.blockchain.models import Transaction
from hub20.apps.raiden.models import (
    ChannelDepositOrder,
    ChannelWithdrawOrder,
    JoinTokenNetworkOrder,
    LeaveTokenNetworkOrder,
    Payment,
    RaidenManagementOrderResult,
    TokenNetworkChannelEvent,
    TokenNetworkChannelStatus,
    UserDepositContractOrder,
)
from hub20.apps.raiden.signals import (
    raiden_payment_received,
    raiden_payment_sent,
    service_deposit_sent,
)

from . import tasks

logger = logging.getLogger(__name__)


@receiver(post_save, sender=Payment)
def on_payment_created_check_received(sender, **kw):
    payment = kw["instance"]
    if kw["created"]:
        if payment.receiver_address == payment.channel.raiden.address:
            logger.info(f"New payment received by {payment.channel}")
            raiden_payment_received.send(sender=Payment, payment=payment)


@receiver(post_save, sender=Payment)
def on_payment_created_check_sent(sender, **kw):
    payment = kw["instance"]
    if kw["created"]:
        if payment.sender_address == payment.channel.raiden.address:
            logger.info(f"New payment sent by {payment.channel}")
            raiden_payment_sent.send(sender=Payment, payment=payment)


@receiver(post_save, sender=TokenNetworkChannelEvent)
def on_token_network_channel_event_set_status(sender, **kw):
    event = kw["instance"]
    if kw["created"]:
        TokenNetworkChannelStatus.set_status(event.channel)


@receiver(post_save, sender=ChannelDepositOrder)
def on_channel_deposit_order_schedule_task(sender, **kw):
    deposit_order = kw["instance"]

    if kw["created"]:
        tasks.make_channel_deposit.delay(order_id=deposit_order.id)


@receiver(post_save, sender=ChannelWithdrawOrder)
def on_channel_withdraw_order_schedule_task(sender, **kw):
    withdraw_order = kw["instance"]

    if kw["created"]:
        tasks.make_channel_withdraw.delay(order_id=withdraw_order.id)


@receiver(post_save, sender=UserDepositContractOrder)
def on_service_deposit_created_schedule_task(sender, **kw):
    deposit_order = kw["instance"]

    if kw["created"]:
        tasks.make_udc_deposit.delay(order_id=deposit_order.id)


@receiver(service_deposit_sent, sender=Transaction)
def on_service_deposit_transaction_create_record(sender, **kw):
    deposit_amount = kw["amount"]
    raiden = kw["raiden"]
    order = UserDepositContractOrder.objects.filter(
        raiden=raiden,
        amount=deposit_amount.amount,
        currency=deposit_amount.currency,
        result__isnull=True,
    ).first()

    if not order:
        return

    RaidenManagementOrderResult.objects.create(transaction=kw["transaction"], order=order)


@receiver(post_save, sender=JoinTokenNetworkOrder)
def on_join_token_network_order_create_schedule_task(sender, **kw):
    order = kw["instance"]

    if kw["created"]:
        tasks.join_token_network.delay(order_id=order.id)


@receiver(post_save, sender=LeaveTokenNetworkOrder)
def on_leave_token_network_order_create_schedule_task(sender, **kw):
    order = kw["instance"]

    if kw["created"]:
        tasks.leave_token_network.delay(order_id=order.id)


__all__ = [
    "on_payment_created_check_received",
    "on_payment_created_check_sent",
    "on_token_network_channel_event_set_status",
    "on_channel_deposit_order_schedule_task",
    "on_channel_withdraw_order_schedule_task",
    "on_service_deposit_created_schedule_task",
    "on_service_deposit_transaction_create_record",
    "on_join_token_network_order_create_schedule_task",
    "on_leave_token_network_order_create_schedule_task",
]



import logging

import celery_pubsub
import httpx
from asgiref.sync import async_to_sync
from celery import shared_task
from channels.layers import get_channel_layer
from django.contrib.auth import get_user_model
from django.contrib.sessions.models import Session
from django.utils import timezone

from .consumers import CheckoutConsumer, Events, SessionEventsConsumer
from .models import Checkout, Transfer

User = get_user_model()
logger = logging.getLogger(__name__)


def _get_open_session_keys():
    now = timezone.now()
    return Session.objects.filter(expire_date__gt=now).values_list("session_key", flat=True)


@shared_task
def execute_transfer(transfer_id):
    try:
        transfer = Transfer.pending.get(id=transfer_id)
        transfer.execute()
    except Transfer.DoesNotExist:
        logger.warning(f"Transfer {transfer_id} not found or already executed")


@shared_task
def execute_pending_transfers():
    for transfer in Transfer.pending.exclude(execute_on__gt=timezone.now()):
        transfer.execute()


@shared_task
def send_session_event(session_key, event, **event_data):
    layer = get_channel_layer()
    channel_group_name = SessionEventsConsumer.get_group_name(session_key)
    event_data.update({"type": "notify_event", "event": event})
    logger.info(f"Sending {event} to session {session_key}")
    async_to_sync(layer.group_send)(channel_group_name, event_data)


@shared_task
def publish_checkout_event(checkout_id, event="checkout.event", **event_data):
    layer = get_channel_layer()
    channel_group_name = CheckoutConsumer.get_group_name(checkout_id)

    logger.info(f"Publishing checkout event {event}. Data: {event_data}")

    event_data.update({"type": "checkout_event", "event_name": event})

    async_to_sync(layer.group_send)(channel_group_name, event_data)


@shared_task
def call_checkout_webhook(checkout_id):
    try:
        checkout = Checkout.objects.get(id=checkout_id)
        url = checkout.store.checkout_webhook_url

        if not url:
            logger.info(f"Checkout {checkout_id} does not have a url")
            return

        try:
            voucher_data = checkout.voucher_data
            voucher_data.update(dict(encoded=checkout.store.issue_jwt(**voucher_data)))
            response = httpx.post(url, json=voucher_data)
            response.raise_for_status()
        except httpx.ConnectError:
            logger.warning(f"Failed to connect to {url}")
        except httpx.HTTPError as exc:
            logger.warning(f"Webhook {url} for {checkout_id} resulted in error response: {exc}")
        except Exception as exc:
            logger.exception(f"Failed to call webhook at {url} for {checkout_id}: {exc}")
    except Checkout.DoesNotExist:
        logger.info(f"Checkout {checkout_id} does not exist")


@shared_task
def notify_new_block(chain_id, block_data, provider_url):

    block_number = block_data["number"]
    session_keys = _get_open_session_keys()
    logger.info(f"Notifying {len(session_keys)} clients about block #{block_number}")
    for session_key in session_keys:
        event_data = dict(
            chain_id=chain_id,
            hash=block_data.hash.hex(),
            number=block_number,
            timestamp=block_data.timestamp,
        )
        send_session_event(session_key, event=Events.BLOCKCHAIN_BLOCK_CREATED.value, **event_data)


@shared_task
def notify_node_unavailable(chain_id, provider_url):
    for session_key in _get_open_session_keys():
        send_session_event(
            session_key, event=Events.ETHEREUM_NODE_UNAVAILABLE.value, chain_id=chain_id
        )


@shared_task
def notify_node_recovered(chain_id, provider_url):
    for session_key in _get_open_session_keys():
        send_session_event(session_key, event=Events.ETHEREUM_NODE_OK.value, chain_id=chain_id)


@shared_task
def clear_expired_sessions():
    Session.objects.filter(expire_date__lte=timezone.now()).delete()


celery_pubsub.subscribe("blockchain.mined.block", notify_new_block)
celery_pubsub.subscribe("node.sync.nok", notify_node_unavailable)
celery_pubsub.subscribe("node.sync.ok", notify_node_recovered)
celery_pubsub.subscribe("node.connection.nok", notify_node_unavailable)
celery_pubsub.subscribe("node.connection.ok", notify_node_recovered)



from datetime import datetime
from typing import Dict, List, Optional

from jsonschema import validate
from pydantic import BaseModel

TOKENLIST_JSON_SCHEMA = {
    "$schema": "http://json-schema.org/draft-07/schema#",
    "$id": "https://uniswap.org/tokenlist.schema.json",
    "title": "Uniswap Token List",
    "description": "Schema for lists of tokens compatible with the Uniswap Interface",
    "definitions": {
        "Version": {
            "type": "object",
            "description": "The version of the list, used in change detection",
            "examples": [{"major": 1, "minor": 0, "patch": 0}],
            "additionalProperties": False,
            "properties": {
                "major": {
                    "type": "integer",
                    "description": "The major version of the list. Must be incremented when tokens are removed from the list or token addresses are changed.",
                    "minimum": 0,
                    "examples": [1, 2],
                },
                "minor": {
                    "type": "integer",
                    "description": "The minor version of the list. Must be incremented when tokens are added to the list.",
                    "minimum": 0,
                    "examples": [0, 1],
                },
                "patch": {
                    "type": "integer",
                    "description": "The patch version of the list. Must be incremented for any changes to the list.",
                    "minimum": 0,
                    "examples": [0, 1],
                },
            },
            "required": ["major", "minor", "patch"],
        },
        "TagIdentifier": {
            "type": "string",
            "description": "The unique identifier of a tag",
            "minLength": 1,
            "maxLength": 10,
            "pattern": "^[\\w]+$",
            "examples": ["compound", "stablecoin"],
        },
        "ExtensionIdentifier": {
            "type": "string",
            "description": "The name of a token extension property",
            "minLength": 1,
            "maxLength": 40,
            "pattern": "^[\\w]+$",
            "examples": ["color", "is_fee_on_transfer", "aliases"],
        },
        "ExtensionMap": {
            "type": "object",
            "description": "An object containing any arbitrary or vendor-specific token metadata",
            "maxProperties": 10,
            "propertyNames": {"$ref": "#/definitions/ExtensionIdentifier"},
            "additionalProperties": {"$ref": "#/definitions/ExtensionValue"},
            "examples": [
                {"color": "#000000", "is_verified_by_me": True},
                {
                    "x-bridged-addresses-by-chain": {
                        "1": {
                            "bridgeAddress": "0x4200000000000000000000000000000000000010",
                            "tokenAddress": "0x4200000000000000000000000000000000000010",
                        }
                    }
                },
            ],
        },
        "ExtensionPrimitiveValue": {
            "anyOf": [
                {"type": "string", "minLength": 1, "maxLength": 42, "examples": ["#00000"]},
                {"type": "boolean", "examples": [True]},
                {"type": "number", "examples": [15]},
                {"type": "null"},
            ]
        },
        "ExtensionValue": {
            "anyOf": [
                {"$ref": "#/definitions/ExtensionPrimitiveValue"},
                {
                    "type": "object",
                    "maxProperties": 10,
                    "propertyNames": {"$ref": "#/definitions/ExtensionIdentifier"},
                    "additionalProperties": {"$ref": "#/definitions/ExtensionValueInner0"},
                },
            ]
        },
        "ExtensionValueInner0": {
            "anyOf": [
                {"$ref": "#/definitions/ExtensionPrimitiveValue"},
                {
                    "type": "object",
                    "maxProperties": 10,
                    "propertyNames": {"$ref": "#/definitions/ExtensionIdentifier"},
                    "additionalProperties": {"$ref": "#/definitions/ExtensionValueInner1"},
                },
            ]
        },
        "ExtensionValueInner1": {"anyOf": [{"$ref": "#/definitions/ExtensionPrimitiveValue"}]},
        "TagDefinition": {
            "type": "object",
            "description": "Definition of a tag that can be associated with a token via its identifier",
            "additionalProperties": False,
            "properties": {
                "name": {
                    "type": "string",
                    "description": "The name of the tag",
                    "pattern": "^[ \\w]+$",
                    "minLength": 1,
                    "maxLength": 20,
                },
                "description": {
                    "type": "string",
                    "description": "A user-friendly description of the tag",
                    "pattern": "^[ \\w\\.,:]+$",
                    "minLength": 1,
                    "maxLength": 200,
                },
            },
            "required": ["name", "description"],
            "examples": [
                {"name": "Stablecoin", "description": "A token with value pegged to another asset"}
            ],
        },
        "TokenInfo": {
            "type": "object",
            "description": "Metadata for a single token in a token list",
            "additionalProperties": False,
            "properties": {
                "chainId": {
                    "type": "integer",
                    "description": "The chain ID of the Ethereum network where this token is deployed",
                    "minimum": 1,
                    "examples": [1, 42],
                },
                "address": {
                    "type": "string",
                    "description": "The checksummed address of the token on the specified chain ID",
                    "pattern": "^0x[a-fA-F0-9]{40}$",
                    "examples": ["0xA0b86991c6218b36c1d19D4a2e9Eb0cE3606eB48"],
                },
                "decimals": {
                    "type": "integer",
                    "description": "The number of decimals for the token balance",
                    "minimum": 0,
                    "maximum": 255,
                    "examples": [18],
                },
                "name": {
                    "type": "string",
                    "description": "The name of the token",
                    "minLength": 1,
                    "maxLength": 40,
                    "pattern": "^[ \\w.'+\\-%/À-ÖØ-öø-ÿ:&\\[\\]\\(\\)]+$",
                    "examples": ["USD Coin"],
                },
                "symbol": {
                    "type": "string",
                    "description": "The symbol for the token; must be alphanumeric",
                    "pattern": "^[a-zA-Z0-9+\\-%/$.]+$",
                    "minLength": 1,
                    "maxLength": 20,
                    "examples": ["USDC"],
                },
                "logoURI": {
                    "type": "string",
                    "description": "A URI to the token logo asset; if not set, interface will attempt to find a logo based on the token address; suggest SVG or PNG of size 64x64",
                    "format": "uri",
                    "examples": ["ipfs://QmXfzKRvjZz3u5JRgC4v5mGVbm9ahrUiB4DgzHBsnWbTMM"],
                },
                "tags": {
                    "type": "array",
                    "description": "An array of tag identifiers associated with the token; tags are defined at the list level",
                    "items": {"$ref": "#/definitions/TagIdentifier"},
                    "maxItems": 10,
                    "examples": ["stablecoin", "compound"],
                },
                "extensions": {"$ref": "#/definitions/ExtensionMap"},
            },
            "required": ["chainId", "address", "decimals", "name", "symbol"],
        },
    },
    "type": "object",
    "additionalProperties": False,
    "properties": {
        "name": {
            "type": "string",
            "description": "The name of the token list",
            "minLength": 1,
            "maxLength": 20,
            "pattern": "^[\\w ]+$",
            "examples": ["My Token List"],
        },
        "timestamp": {
            "type": "string",
            "format": "date-time",
            "description": "The timestamp of this list version; i.e. when this immutable version of the list was created",
        },
        "version": {"$ref": "#/definitions/Version"},
        "tokens": {
            "type": "array",
            "description": "The list of tokens included in the list",
            "items": {"$ref": "#/definitions/TokenInfo"},
            "minItems": 1,
            "maxItems": 10000,
        },
        "keywords": {
            "type": "array",
            "description": "Keywords associated with the contents of the list; may be used in list discoverability",
            "items": {
                "type": "string",
                "description": "A keyword to describe the contents of the list",
                "minLength": 1,
                "maxLength": 20,
                "pattern": "^[\\w ]+$",
                "examples": ["compound", "lending", "personal tokens"],
            },
            "maxItems": 20,
            "uniqueItems": True,
        },
        "tags": {
            "type": "object",
            "description": "A mapping of tag identifiers to their name and description",
            "propertyNames": {"$ref": "#/definitions/TagIdentifier"},
            "additionalProperties": {"$ref": "#/definitions/TagDefinition"},
            "maxProperties": 20,
            "examples": [
                {
                    "stablecoin": {
                        "name": "Stablecoin",
                        "description": "A token with value pegged to another asset",
                    }
                }
            ],
        },
        "logoURI": {
            "type": "string",
            "description": "A URI for the logo of the token list; prefer SVG or PNG of size 256x256",
            "format": "uri",
            "examples": ["ipfs://QmXfzKRvjZz3u5JRgC4v5mGVbm9ahrUiB4DgzHBsnWbTMM"],
        },
    },
    "required": ["name", "timestamp", "version", "tokens"],
}


class TokenListVersionModel(BaseModel):
    major: int
    minor: int
    patch: int

    @property
    def as_string(self):
        return f"{self.major}.{self.minor}.{self.patch}"


class TokenListTagDefinition(BaseModel):
    name: Optional[str]
    description: Optional[str]


class TokenInfo(BaseModel):
    chainId: int
    address: str
    decimals: int
    name: str
    symbol: str
    logoURI: Optional[str]
    tags: Optional[List[str]]


class TokenList(BaseModel):
    name: str
    timestamp: datetime
    version: TokenListVersionModel
    logoURI: str
    tokens: List[TokenInfo]
    tags: Optional[Dict[str, TokenListTagDefinition]]
    keywords: List[str]


def validate_token_list(token_list_data):
    return validate(instance=token_list_data, schema=TOKENLIST_JSON_SCHEMA)



import getpass
import logging

import ethereum
from django.core.management.base import BaseCommand
from eth_utils import to_checksum_address

from hub20.apps.blockchain.client import make_web3
from hub20.apps.blockchain.models import Web3Provider
from hub20.apps.ethereum_money.models import EthereumTokenAmount, KeystoreAccount
from hub20.apps.raiden.client.blockchain import get_service_token, make_service_deposit

logger = logging.getLogger(__name__)


class Command(BaseCommand):
    help = "Deposits RDN at Raiden Service Contract"

    def add_arguments(self, parser):
        parser.add_argument("-a", "--account", required=True, type=str)
        parser.add_argument("--amount", default=1000, type=int)
        parser.add_argument("--chain-id", "-c", dest="chain_id", required=True, type=int)

    def handle(self, *args, **options):
        address = to_checksum_address(options["account"])
        account = KeystoreAccount.objects.filter(address=address).first()

        if not account:
            private_key = getpass.getpass(f"Private Key for {address} required: ")
            generated_address = to_checksum_address(ethereum.utils.privtoaddr(private_key.strip()))
            assert generated_address == address, "Private Key does not match"
            account = KeystoreAccount(address=address, private_key=private_key)

        provider = Web3Provider.available.get(chain_id=options["chain_id"])
        w3 = make_web3(provider=provider)

        token = get_service_token(w3=w3)

        amount = EthereumTokenAmount(amount=options["amount"], currency=token)
        make_service_deposit(w3=w3, account=account, amount=amount)



from channels.routing import ProtocolTypeRouter, URLRouter
from django import setup
from django.core.asgi import get_asgi_application
from django.urls import path

setup()

from hub20.apps.core.api import consumer_patterns  # isort:skip
from hub20.apps.core.middleware import TokenAuthMiddlewareStack  # isort:skip

application = ProtocolTypeRouter(
    {
        "http": get_asgi_application(),
        "websocket": TokenAuthMiddlewareStack(
            URLRouter(
                [
                    path("ws/", URLRouter(consumer_patterns)),
                ]
            )
        ),
    }
)



from __future__ import annotations

import datetime
from typing import Optional

from django.contrib.auth import get_user_model
from django.contrib.postgres.fields import ArrayField
from django.db import models
from django.db.models import F, Max
from ethereum.utils import checksum_encode
from model_utils.choices import Choices
from model_utils.managers import InheritanceManager, QueryManager
from model_utils.models import StatusModel, TimeStampedModel

from hub20.apps.blockchain.fields import EthereumAddressField, HexField, Uint256Field
from hub20.apps.blockchain.models import BaseEthereumAccount, Chain, Transaction, Web3Provider
from hub20.apps.blockchain.typing import Address
from hub20.apps.blockchain.validators import uri_parsable_scheme_validator
from hub20.apps.ethereum_money.models import (
    EthereumToken,
    EthereumTokenAmount,
    EthereumTokenAmountField,
    EthereumTokenValueModel,
)

CHANNEL_STATUSES = Choices("opened", "settling", "settled", "unusable", "closed", "closing")
User = get_user_model()

raiden_url_validator = uri_parsable_scheme_validator(("http", "https"))


class RaidenURLField(models.URLField):
    default_validators = [raiden_url_validator]


class TokenNetwork(models.Model):
    address = EthereumAddressField()
    token = models.OneToOneField(EthereumToken, on_delete=models.CASCADE)
    objects = models.Manager()

    def can_reach(self, address) -> bool:
        # This is a very naive assumption. One should not assume that we can
        # reach an address just because the address has an open channel.

        # However, our main purpose is only to find out if a given address is
        # being used by raiden and that we can _try_ to use for a transfer.
        open_channels = self.channels.filter(status__status=CHANNEL_STATUSES.opened)
        return open_channels.filter(participant_addresses__contains=[address]).exists()

    @property
    def chain(self):
        return self.token.chain

    @property
    def events(self):
        return TokenNetworkChannelEvent.objects.filter(channel__token_network=self)

    @property
    def most_recent_channel_event(self) -> Optional[int]:
        max_block_aggregate = Max("tokennetworkchannelevent__transaction__block__number")
        return self.channels.aggregate(max_block=max_block_aggregate).get("max_block")

    def __str__(self):
        return f"{self.address} - ({self.token.symbol} @ {self.token.chain_id})"


class TokenNetworkChannel(models.Model):
    token_network = models.ForeignKey(
        TokenNetwork, on_delete=models.CASCADE, related_name="channels"
    )
    identifier = Uint256Field()
    participant_addresses = ArrayField(EthereumAddressField(), size=2)

    @property
    def events(self):
        return self.tokennetworkchannelevent_set.order_by("transaction__block__number")


class TokenNetworkChannelStatus(StatusModel):
    STATUS = CHANNEL_STATUSES
    channel = models.OneToOneField(
        TokenNetworkChannel, on_delete=models.CASCADE, related_name="status"
    )

    @classmethod
    def set_status(cls, channel: TokenNetworkChannel):
        last_event = channel.events.last()
        event_name = last_event and last_event.name
        status = (
            event_name
            and {
                "ChannelOpened": CHANNEL_STATUSES.opened,
                "ChannelClosed": CHANNEL_STATUSES.closed,
            }.get(event_name)
        )
        if status:
            cls.objects.update_or_create(channel=channel, defaults={"status": status})


class TokenNetworkChannelEvent(models.Model):
    channel = models.ForeignKey(TokenNetworkChannel, on_delete=models.CASCADE)
    transaction = models.ForeignKey(Transaction, on_delete=models.CASCADE)
    name = models.CharField(max_length=32, db_index=True)

    class Meta:
        unique_together = ("channel", "transaction")


class Raiden(BaseEthereumAccount):
    url = RaidenURLField(unique=True)
    web3_provider = models.ForeignKey(Web3Provider, null=True, on_delete=models.SET_NULL)

    @property
    def private_key(self):
        return None

    @property
    def token_networks(self):
        return TokenNetwork.objects.filter(
            channel__status=Channel.STATUS.opened, channel__raiden=self
        )

    @property
    def open_channels(self):
        return self.channels.filter(status=Channel.STATUS.opened)

    @property
    def chain(self) -> Optional[Chain]:
        return self.web3_provider and self.web3_provider.chain

    @property
    def payments(self):
        return Payment.objects.filter(channel__raiden=self)

    @property
    def payments_received(self):
        return Payment.received.filter(channel__raiden=self)

    @property
    def payments_sent(self):
        return Payment.sent.filter(channel__raiden=self)

    @classmethod
    def generate(cls, address: Address, url: str):
        raiden, _ = cls.objects.get_or_create(address=checksum_encode(address).hex(), url=url)
        return raiden

    def __str__(self):
        return f"Raiden @ {self.url} ({self.address})"


class Channel(StatusModel):
    STATUS = CHANNEL_STATUSES
    raiden = models.ForeignKey(Raiden, on_delete=models.CASCADE, related_name="channels")
    token_network = models.ForeignKey(TokenNetwork, on_delete=models.CASCADE)
    identifier = models.PositiveIntegerField()
    partner_address = EthereumAddressField(db_index=True)
    balance = EthereumTokenAmountField()
    total_deposit = EthereumTokenAmountField()
    total_withdraw = EthereumTokenAmountField()

    objects = models.Manager()
    funded = QueryManager(status=STATUS.opened, balance__gt=0)
    available = QueryManager(status=STATUS.opened)

    @property
    def token(self):
        return self.token_network.token

    @property
    def balance_amount(self) -> EthereumTokenAmount:
        return EthereumTokenAmount(amount=self.balance, currency=self.token)

    @property
    def deposit_amount(self) -> EthereumTokenAmount:
        return EthereumTokenAmount(amount=self.total_deposit, currency=self.token)

    @property
    def withdraw_amount(self) -> EthereumTokenAmount:
        return EthereumTokenAmount(amount=self.total_withdraw, currency=self.token)

    @property
    def last_event_timestamp(self) -> Optional[datetime.datetime]:
        latest_event = self.payments.order_by("-timestamp").first()
        return latest_event and latest_event.timestamp

    def __str__(self):
        return f"Channel {self.identifier} ({self.partner_address})"

    @classmethod
    def make(cls, raiden: Raiden, **channel_data) -> Optional[Channel]:
        token_network_address = channel_data.pop("token_network_address")
        token_address = channel_data.pop("token_address")

        token = EthereumToken.ERC20tokens.filter(address=token_address).first()

        assert token is not None
        assert token.chain == raiden.chain

        token_network, _ = TokenNetwork.objects.get_or_create(
            address=token_network_address, token=token
        )

        assert token_network.token.address == token_address

        balance = token.from_wei(channel_data.pop("balance"))
        total_deposit = token.from_wei(channel_data.pop("total_deposit"))
        total_withdraw = token.from_wei(channel_data.pop("total_withdraw"))

        channel, _ = raiden.channels.update_or_create(
            identifier=channel_data["channel_identifier"],
            partner_address=channel_data["partner_address"],
            token_network=token_network,
            defaults={
                "status": channel_data["state"],
                "balance": balance.amount,
                "total_deposit": total_deposit.amount,
                "total_withdraw": total_withdraw.amount,
            },
        )
        return channel

    class Meta:
        unique_together = (
            ("raiden", "token_network", "partner_address"),
            ("raiden", "token_network", "identifier"),
        )


class Payment(models.Model):
    MAX_IDENTIFIER_ID = (2 ** 64) - 1
    channel = models.ForeignKey(Channel, on_delete=models.CASCADE, related_name="payments")
    amount = EthereumTokenAmountField()
    timestamp = models.DateTimeField()
    identifier = Uint256Field()
    sender_address = EthereumAddressField()
    receiver_address = EthereumAddressField()
    objects = models.Manager()
    sent = QueryManager(sender_address=F("channel__raiden__address"))
    received = QueryManager(receiver_address=F("channel__raiden__address"))

    @property
    def token(self):
        return self.channel.token

    @property
    def as_token_amount(self) -> EthereumTokenAmount:
        return EthereumTokenAmount(amount=self.amount, currency=self.token)

    @property
    def partner_address(self):
        return self.receiver_address if self.is_outgoing else self.sender_address

    @property
    def is_outgoing(self):
        return self.channel.raiden.address == self.sender_address

    @property
    def is_incoming(self):
        return self.channel.raiden.address == self.receiver_address

    @classmethod
    def make(cls, channel: Channel, **payment_data):
        sender_address = payment_data.pop("sender_address")
        receiver_address = payment_data.pop("receiver_address")
        identifier = payment_data.pop("identifier")

        payment, _ = channel.payments.update_or_create(
            channel=channel,
            sender_address=sender_address,
            receiver_address=receiver_address,
            identifier=identifier,
            defaults=payment_data,
        )
        return payment

    class Meta:
        unique_together = ("channel", "identifier", "sender_address", "receiver_address")


class RaidenManagementOrder(TimeStampedModel):
    raiden = models.ForeignKey(Raiden, on_delete=models.CASCADE)
    user = models.ForeignKey(User, on_delete=models.PROTECT)
    objects = InheritanceManager()
    transaction_hash = HexField(max_length=64, unique=True, null=True)


class JoinTokenNetworkOrder(RaidenManagementOrder):
    token_network = models.ForeignKey(TokenNetwork, on_delete=models.PROTECT)
    amount = EthereumTokenAmountField()


class LeaveTokenNetworkOrder(RaidenManagementOrder):
    token_network = models.ForeignKey(TokenNetwork, on_delete=models.PROTECT)


class ChannelDepositOrder(RaidenManagementOrder):
    channel = models.ForeignKey(Channel, on_delete=models.PROTECT)
    amount = EthereumTokenAmountField()


class ChannelWithdrawOrder(RaidenManagementOrder):
    channel = models.ForeignKey(Channel, on_delete=models.PROTECT)
    amount = EthereumTokenAmountField()


class UserDepositContractOrder(RaidenManagementOrder, EthereumTokenValueModel):
    chain = models.ForeignKey(Chain, related_name="user_deposit_orders", on_delete=models.CASCADE)


class RaidenManagementOrderResult(TimeStampedModel):
    order = models.OneToOneField(
        RaidenManagementOrder, on_delete=models.CASCADE, related_name="result"
    )
    transaction = models.OneToOneField(Transaction, null=True, on_delete=models.SET_NULL)


class RaidenManagementOrderError(TimeStampedModel):
    order = models.OneToOneField(
        RaidenManagementOrder, on_delete=models.CASCADE, related_name="error"
    )
    message = models.TextField(null=True, blank=True)
    transaction = models.OneToOneField(Transaction, null=True, on_delete=models.SET_NULL)


__all__ = [
    "TokenNetwork",
    "Raiden",
    "TokenNetwork",
    "Channel",
    "Payment",
    "RaidenManagementOrder",
    "JoinTokenNetworkOrder",
    "LeaveTokenNetworkOrder",
    "ChannelDepositOrder",
    "ChannelWithdrawOrder",
    "UserDepositContractOrder",
    "RaidenManagementOrderResult",
    "RaidenManagementOrderError",
]



import logging

from django.core.management.base import BaseCommand

from hub20.apps.ethereum_money.models import TokenList

logger = logging.getLogger(__name__)


class Command(BaseCommand):
    help = "Loads token lists (https://tokenlists.org/) according to the schema"

    def add_arguments(self, parser):
        parser.add_argument("urls", metavar="N", nargs="+", type=str)

    def handle(self, *args, **options):

        for url in options["urls"]:
            try:
                token_list_data = TokenList.fetch(url)
                TokenList.make(url, token_list_data=token_list_data)
            except Exception as exc:
                logger.info(f"Failed to make list from {url}: {exc}")



# Generated by Django 4.0 on 2022-01-28 23:27

import django.contrib.postgres.fields
import django.contrib.postgres.fields.hstore
import django.db.models.deletion
from django.contrib.postgres.operations import HStoreExtension
from django.db import migrations, models

import hub20.apps.blockchain.fields


class Migration(migrations.Migration):

    initial = True

    dependencies = []

    operations = [
        HStoreExtension(),
        migrations.CreateModel(
            name="Block",
            fields=[
                (
                    "hash",
                    hub20.apps.blockchain.fields.HexField(
                        max_length=64, primary_key=True, serialize=False
                    ),
                ),
                ("number", models.PositiveIntegerField(db_index=True)),
                ("base_fee_per_gas", models.PositiveBigIntegerField(null=True)),
                ("timestamp", models.DateTimeField()),
                ("parent_hash", hub20.apps.blockchain.fields.HexField(max_length=64)),
                (
                    "uncle_hashes",
                    django.contrib.postgres.fields.ArrayField(
                        base_field=hub20.apps.blockchain.fields.HexField(max_length=64), size=None
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="Chain",
            fields=[
                ("id", models.PositiveBigIntegerField(primary_key=True, serialize=False)),
                ("name", models.CharField(default="EVM-compatible network", max_length=128)),
                ("is_mainnet", models.BooleanField(default=True)),
                ("highest_block", models.PositiveIntegerField()),
            ],
            options={
                "ordering": ("id",),
            },
        ),
        migrations.CreateModel(
            name="Web3Provider",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True, primary_key=True, serialize=False, verbose_name="ID"
                    ),
                ),
                ("url", hub20.apps.blockchain.fields.Web3ProviderURLField()),
                ("client_version", models.CharField(max_length=300, null=True)),
                ("requires_geth_poa_middleware", models.BooleanField(default=False)),
                ("supports_pending_filters", models.BooleanField(default=False)),
                ("supports_eip1559", models.BooleanField(default=False)),
                ("is_active", models.BooleanField(default=True)),
                ("synced", models.BooleanField(default=False)),
                ("connected", models.BooleanField(default=False)),
                (
                    "chain",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="providers",
                        to="blockchain.chain",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="Transaction",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True, primary_key=True, serialize=False, verbose_name="ID"
                    ),
                ),
                ("hash", hub20.apps.blockchain.fields.HexField(db_index=True, max_length=64)),
                ("from_address", hub20.apps.blockchain.fields.EthereumAddressField(db_index=True)),
                ("to_address", hub20.apps.blockchain.fields.EthereumAddressField(db_index=True)),
                ("receipt", django.contrib.postgres.fields.hstore.HStoreField()),
                (
                    "block",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="transactions",
                        to="blockchain.block",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="NativeToken",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True, primary_key=True, serialize=False, verbose_name="ID"
                    ),
                ),
                ("name", models.CharField(max_length=500)),
                ("symbol", models.CharField(max_length=16)),
                ("decimals", models.PositiveIntegerField(default=18)),
                (
                    "chain",
                    models.OneToOneField(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="native_token",
                        to="blockchain.chain",
                    ),
                ),
            ],
            options={
                "abstract": False,
            },
        ),
        migrations.CreateModel(
            name="Explorer",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True, primary_key=True, serialize=False, verbose_name="ID"
                    ),
                ),
                ("name", models.CharField(max_length=200, null=True)),
                ("url", models.URLField()),
                ("standard", models.CharField(max_length=200, null=True)),
                (
                    "chain",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="explorers",
                        to="blockchain.chain",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="EventIndexer",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True, primary_key=True, serialize=False, verbose_name="ID"
                    ),
                ),
                ("name", models.CharField(max_length=200)),
                ("last_block", models.PositiveBigIntegerField(default=1)),
                (
                    "chain",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="indexers",
                        to="blockchain.chain",
                    ),
                ),
            ],
        ),
        migrations.AddField(
            model_name="block",
            name="chain",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="blocks",
                to="blockchain.chain",
            ),
        ),
        migrations.CreateModel(
            name="BaseEthereumAccount",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True, primary_key=True, serialize=False, verbose_name="ID"
                    ),
                ),
                (
                    "address",
                    hub20.apps.blockchain.fields.EthereumAddressField(db_index=True, unique=True),
                ),
                ("transactions", models.ManyToManyField(to="blockchain.Transaction")),
            ],
        ),
        migrations.CreateModel(
            name="TransactionDataRecord",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True, primary_key=True, serialize=False, verbose_name="ID"
                    ),
                ),
                ("hash", hub20.apps.blockchain.fields.HexField(db_index=True, max_length=64)),
                ("from_address", hub20.apps.blockchain.fields.EthereumAddressField(db_index=True)),
                ("to_address", hub20.apps.blockchain.fields.EthereumAddressField(db_index=True)),
                ("data", django.contrib.postgres.fields.hstore.HStoreField()),
                (
                    "chain",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="transaction_data",
                        to="blockchain.chain",
                    ),
                ),
            ],
            options={
                "unique_together": {("hash", "chain")},
            },
        ),
        migrations.AddConstraint(
            model_name="transaction",
            constraint=models.UniqueConstraint(
                fields=("hash", "block"), name="unique_tx_per_block"
            ),
        ),
        migrations.AlterUniqueTogether(
            name="explorer",
            unique_together={("url", "chain")},
        ),
        migrations.AddConstraint(
            model_name="eventindexer",
            constraint=models.UniqueConstraint(
                fields=("chain", "name"), name="unique_name_per_chain"
            ),
        ),
        migrations.AlterUniqueTogether(
            name="block",
            unique_together={("chain", "hash", "number")},
        ),
    ]



from django.contrib.auth import get_user_model
from django.db.models import Q
from django_filters import rest_framework as filters

User = get_user_model()

from .models import Deposit


class UserFilter(filters.FilterSet):
    search = filters.CharFilter(label="search", method="user_suggestion")

    def user_suggestion(self, queryset, name, value):
        q_username = Q(username__istartswith=value)
        q_first_name = Q(first_name__istartswith=value)
        q_last_name = Q(last_name__istartswith=value)
        q_email = Q(email__istartswith=value)
        return queryset.filter(q_username | q_first_name | q_last_name | q_email)

    class Meta:
        model = User
        fields = ("search",)


class DepositFilter(filters.FilterSet):
    open = filters.BooleanFilter(label="open", method="filter_open")
    chain = filters.NumberFilter(field_name="currency__chain")
    token = filters.CharFilter(field_name="currency__address")

    def filter_open(self, queryset, name, value):
        return queryset.open() if value else queryset.expired()

    class Meta:
        model = Deposit
        fields = ("token", "chain", "open")



from django.apps import apps as django_apps
from django.conf import settings
from django.core.exceptions import ImproperlyConfigured


def get_wallet_model():
    """
    Return the Wallet model that is active in this project.
    """

    account_setting = getattr(settings, "WALLET_MODEL", "wallet.ColdWallet")

    try:
        return django_apps.get_model(account_setting, require_ready=False)
    except ValueError:
        raise ImproperlyConfigured("WALLET_MODEL must be of the form 'app_label.model_name'")
    except LookupError:
        raise ImproperlyConfigured(
            f"Model '{account_setting}' is not installed and can not be used as Ethereum Account"
        )



import logging

from django.core.management.base import BaseCommand
from django.db import IntegrityError

from hub20.apps.blockchain.client import get_web3
from hub20.apps.blockchain.models import Chain, Web3Provider

logger = logging.getLogger(__name__)


class Command(BaseCommand):
    help = "Saves a new blockchain into the database"

    def add_arguments(self, parser):
        parser.add_argument("--id", dest="chain_id", required=True, type=int)
        parser.add_argument("--name", dest="name", required=True, type=str)
        parser.add_argument("--provider-url", dest="provider", required=False, type=str)

    def handle(self, *args, **options):
        chain_id = options["chain_id"]
        name = options["name"]
        try:
            Chain.objects.create(id=chain_id, name=name, highest_block=0)
        except IntegrityError:
            logger.warning(f"Chain {chain_id} already exists. Aborting...")
            return

        provider_url = options["provider_url"]
        if provider_url:
            w3 = get_web3(provider_url=provider_url)

            if w3.eth.chain_id != chain_id:
                logger.warning(
                    f"Provider reported chain #{w3.eth.chain_id}, expected {chain_id}. Skipping..."
                )
            else:
                provider, _ = Web3Provider.objects.get_or_create(
                    chain_id=chain_id, url=provider_url
                )
                provider.activate()
